from bioservices.ena import ENA
from bioservices.eutils import EUtils


def download_gbk(accession, output_filename=None, method="EUtils", service=None):
    """Utility to download a GFF3 file from ENA or EUtils

    :param accession: a valid accession number with possible version (see
        example)
    :param output_filename: if none, use accession + fa extension (replaces dot
        with underscore)
    :param method: either EUtils or ENA
    :param service: an existing instance of ENA or EUtils. This is useful to
        call this functions many times. The creation of the service is indeed
        time consuming. If used, then **method** is ignored.

    ::

        download_gbk("FN433596.1")

    """
    if service:
        method = service.services.name

    if output_filename is None:
        output_filename = accession.replace(".", "_") + ".gbk"

    if method == "EUtils":
        _download_gbk_ncbi(accession, output_filename, service)
    elif method == "ENA":
        _download_gbk_ena(accession, output_filename, service)
    else:
        raise ValueError("method or service must be either ENA or EUtils")


def _download_gbk_ena(accession, output_filename, service=None):
    raise NotImplementedError


def _download_gbk_ncbi(accession, output_filename, service=None):
    if service is None:
        service = EUtils()
    data = service.EFetch("nucleotide", accession, rettype="gb")
    data = data.decode()
    # Save to local file
    with open(output_filename, "w") as fout:
        fout.write(data)





#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to some part of the UniProt web service

.. topic:: What is COG service?

    :URL: https://www.ncbi.nlm.nih.gov/research/cog/webservices/
    :Citation:

    .. highlights::

        Database of Clusters of Orthologous Genes (COGs)

        -- From COG web site, Jan 2021


"""
import types
import io
import sys

from bioservices.services import REST
from bioservices import logger

logger.name = __name__

try:
    import pandas as pd
except:
    pass


__all__ = ["COG"]


class COG:
    """Interface to the COG service


    from bioservices import COG
    c = COG()
    cogs = c.get_all_cogs()   # This is a pandas dataframe

    """

    _url = "https://www.ncbi.nlm.nih.gov/research/cog/api"

    def __init__(self, verbose=False, cache=False):
        """**Constructor**"""
        self.services = REST(name="cog", url=COG._url, verbose=verbose, cache=cache)

    def get_cogs(self, page=1):
        """Get COGs. Unfortunately, the API sends 10 COGS at a tine given a
        specific page.

        The dictionary returned contains the results, count, previous and next
        page.
        """
        res = self.services.http_get("cog", frmt="json", params={"page": page})
        return res

    def get_cogs_by_gene(self, gene):
        """Filter COGs by gene tag: MK0280"""
        res = self.services.http_get("cog", frmt="json", params={"gene": gene})
        return res

    def get_cogs_by_id(self, cog_id):
        """Filter COGs by COG ID tag: COG0003"""
        res = self.services.http_get("cog", frmt="json", params={"cog": cog_id})
        return res

    def get_cogs_by_assembly_id(self, assembly_id):
        """Filter COGs by assembly ID: GCA_000007185.1"""
        res = self.services.http_get("cog", frmt="json", params={"assembly": assembly_id})
        return res

    def get_cogs_by_orgnanism(self, name):
        """Filter COGs by organism name: Nitrosopumilus_maritimus_SCM1"""
        res = self.services.http_get("cog", frmt="json", params={"organism": name})
        return res

    def get_cogs_by_taxon_id(self, taxon_id):
        """Filter COGs by taxid: 1229908"""
        res = self.services.http_get("cog", frmt="json", params={"taxid": taxon_id})
        return res

    def get_cogs_by_category(self, category):
        """Filter COGs by Taxonomic Category: ACTINOBACTERIA"""
        res = self.services.http_get("cog", frmt="json", params={"category": category})
        return res

    def get_cogs_by_category_id(self, category):
        """Filter COGs by Taxonomic Category taxid: 651137"""
        res = self.services.http_get("cog", frmt="json", params={"cat_taxid": category})
        return res

    def get_cogs_by_category_(self, protein):
        """Filter COGs by Protein name: AJP49128.1"""
        res = self.services.http_get("cog", frmt="json", params={"protein": protein})
        return res

    # The search keywords (cog, assembly, organism, taxid, category, cat_taxid and protein)
    # can be combined to filter the COG lists.

    def get_cogs_by_id_and_category(self, cog_id, category):
        """Filter COGs by COG id and Taxonomy Categories: COG0004 and CYANOBACTERIA"""
        res = self.services.http_get("cog", frmt="json", params={"cog": cog_id, "category": category})
        return res

    def get_cogs_by_id_and_organism(self, cog_id, organism):
        """Filter COGs by COG id and organism: COG0004 and Escherichia_coli_K-12_sub_MG1655"""
        res = self.services.http_get("cog", frmt="json", params={"cog": cog_id, "organism,": organism})
        return res

    def get_all_cogs_definition(self):
        """Get all COG Definitions:"""
        res = self.services.http_get("cogdef", frmt="json")
        return res

    def get_cog_definition_by_cog_id(self, cog_id):
        """Get specific COG Definitions by COG: COG0003"""
        res = self.services.http_get("cogdef", frmt="json", params={"cog": cog_id})
        return res

    def get_cog_definition_by_name(self, cog):
        """Get specific COG Definitions by name: Thiamin-binding stress-response protein YqgV, UPF0045 family"""
        res = self.services.http_get("cogdef", frmt="json", params={"name": cog})
        return res

    def get_taxonomic_categories(self):
        """Get all Taxonomic Categories:"""
        res = self.services.http_get("taxonomy", frmt="json")
        return res

    def get_taxonomic_category_by_name(self, name):
        """Get specific Taxonomic Category by name: ALPHAPROTEOBACTERIA"""
        res = self.services.http_get("taxonomy", frmt="json", params={"name": name})
        return res



try:
    from collections import OrderedDict
except:
    # for python 2.6
    from ordereddict import OrderedDict
from easydev.decorators import ifpylab, ifpandas

__all__ = ["FASTA", "MultiFASTA"]


class MultiFASTA(object):
    """Class to manipulate several several FASTA items

    Here, we load some FASTA using UniProt web service::

        >>> from bioservices import MultiFASTA
        >>> mf = MultiFASTA()
        >>> mf.load_fasta("P43408")
        >>> mf.load_fasta("P21318")

    You can then get back to your accession entries as follows ::

        >>> mf.ids
        ['P43408', 'P21318']

    And the sequences in the same order can be accessed::

        >>> len(mf)
        2

    Each FASTA is stored in :attr:`fasta`, which is a dictionary where each
    values is an instance of :class:`FASTA`::

        >>> print(mf._fasta["P43408"].fasta)
        >sp|P43408|KADA_METIG Adenylate kinase OS=Methanotorris igneus GN=adkA PE=1 SV=2
        MKNKVVVVTGVPGVGGTTLTQKTIEKLKEEGIEYKMVNFGTVMFEVAKEEGLVEDRDQMR
        KLDPDTQKRIQKLAGRKIAEMAKESNVIVDTHSTVKTPKGYLAGLPIWVLEELNPDIIVI
        VETSSDEILMRRLGDATRNRDIELTSDIDEHQFMNRCAAMAYGVLTGATVKIIKNRDGLL
        DKAVEELISVLK

    The most convenient way to access to all data is to use the dataframe attribute::

        >>> mf.df.Sequence


    .. plot::
        :width: 50%
        :include-source:

        >>> from bioservices.apps import MultiFASTA
        >>> f = MultiFASTA()
        >>> f.load_fasta(["P43403", "P43410"])
        >>> f.df.Size.hist()

    """

    def __init__(self):
        # fetch the sequence using this attribute
        self._fasta_fetcher = FASTA()

        # an ordered dictionary to store the fasta contents
        # not python2.6 compatible
        self._fasta = OrderedDict()

    def __len__(self):
        return len(self._fasta)

    def _get_fasta(self):
        return self._fasta

    fasta = property(_get_fasta, doc="Returns all FASTA instances ")

    def _get_ids(self):
        return [f for f in self._fasta.keys()]

    ids = property(_get_ids, doc="returns list of keys/accession identifiers")

    def load_fasta(self, ids):
        """Loads a single FASTA file into the dictionary"""
        if isinstance(ids, str):
            ids = [ids]

        for id_ in ids:
            self._fasta_fetcher.load(id_)
            # create a new instance of FASTA and save fasta data
            f = FASTA()
            f._fasta = self._fasta_fetcher._fasta[:]

            # append in the ordered dictionary
            self._fasta[id_] = f
            print("%s loaded" % id_)

    def save_fasta(self, filename):
        """Save all FASTA into a file"""
        fh = open(filename, "w")
        for f in self._fasta.values():
            fh.write(f.fasta)
        fh.close()

    def read_fasta(self, filename):
        """Load several FASTA from a filename"""
        fh = open(filename, "r")
        data = fh.read()
        fh.close()

        # we split according to ">2 character
        for thisfasta in data.split(">")[1:]:
            f = FASTA()
            f._fasta = f._interpret(thisfasta)
            if f.accession != None and f.accession not in self.ids:
                self._fasta[f.accession] = f
            else:
                print("Accession %s is already in the ids list or could not be interpreted. skipped" % str(f.accession))

    @ifpandas
    def _get_df(self):
        import pandas as pd

        df = pd.concat([self.fasta[id_].df for id_ in self.fasta.keys()])
        df.reset_index(inplace=True)
        return df

    df = property(_get_df)

    @ifpylab
    def hist_size(self, **kargs):
        import pylab

        self.df.Size.hist(**kargs)
        pylab.title("Histogram length of the sequences")
        pylab.xlabel("Length")


class FASTA(object):
    """Dedicated class to manipulates FASTA sequence(s)

    Here is a FASTA file example::

        >sp|P43408|KADA_METIG Adenylate kinase OS=Methanotorris igneus GN=adkA PE=1 SV=2
        MKNKVVVVTGVPGVGGTTLTQKTIEKLKEEGIEYKMVNFGTVMFEVAKEEGLVEDRDQMR
        KLDPDTQKRIQKLAGRKIAEMAKESNVIVDTHSTVKTPKGYLAGLPIWVLEELNPDIIVI
        VETSSDEILMRRLGDATRNRDIELTSDIDEHQFMNRCAAMAYGVLTGATVKIIKNRDGLL
        DKAVEELISVLK

    The format is made of a header and a sequence. Any FASTA can be read
    and the pair of header/sequence retrieved from the :attr:`sequence` and
    :attr:`header` attributes. However, headers differ from one database to
    another one and interpretation is not implemented except for SWISS-PROT.
    Identifiers can be retrieved whatsoever.

    You can read a FASTA sequence from a local file or download one from UniProt

    .. doctest::

        >>> from bioservices.apps.fasta import FASTA
        >>> f = FASTA()
        >>> f.load("P43403")
        >>> acc = f.accession    # the accession (P43403)
        >>> fasta = f.fasta      # raw FASTA string
        >>> seq = f.sequence     # the sequence itself
        >>> header = f.header    # the header itself
        >>> identifier = f.identifier

    You can also get a dataframe also using Pandas library.::

        >>> f.df

    The columns stored in the dataframe encompase:

        * **Accession** that is taken from the header (e.g., P43403 from uniprot)
        * **Sequence**, a copy of the FASTA sequence
        * **Size**,  the length of the sequence.
        * **Database**, the database type found in the header (e.g., sp for
          SWISS-PROT; see below for a list of database and their header format).
        * Some column such as **Organism** are filled only for some database
        * **Identififers** is the begining of the header.

    .. seealso:: :class:`MultiFASTA` for multi FASTA manipulation.
    .. addedversion: 1.2.0 maybe merged with :class:`MulitFASTA` class

    List of identifiers corresponding to different databases.

    ================================= ====================================
    ================================= ====================================
    GenBank                           gi|gi-number|gb|accession|locus
    EMBL Data Library                 gi|gi-number|emb|accession|locus
    DDBJ, DNA Database of Japan       gi|gi-number|dbj|accession|locus
    NBRF PIR                          pir||entry
    Protein Research Foundation       prf||name
    SWISS-PROT                        sp|accession|name
    Brookhaven Protein Data Bank (1)  pdb|entry|chain
    Brookhaven Protein Data Bank (2)  entry:chain|PDBID|CHAIN|SEQUENCE
    Patents                           pat|country|number
    GenInfo Backbone Id               bbs|number
    General database identifier       gnl|database|identifier
    NCBI Reference Sequence           ref|accession|locus
    Local Sequence identifier         lcl|identifier
    ================================= ====================================


    The :meth::`load_fasta` relies on UniProt service.
    """

    known_dbtypes = ["sp", "gi"]

    def __init__(self):
        self._fasta = None

    def _get_fasta(self):
        return self._fasta

    fasta = property(_get_fasta, doc="returns FASTA content")

    # for all types
    def _get_sequence(self):
        if self.fasta:
            return "".join(self.fasta.split("\n")[1:])
        else:
            raise ValueError("You need to load a fasta sequence first using get_fasta or read_fasta")

    sequence = property(_get_sequence, doc="returns the sequence only")

    # for all types
    def _get_header(self):
        if self.fasta:
            return self.fasta.split("\n")[0]
        else:
            raise ValueError("You need to load a fasta sequence first using get_fasta or read_fasta")

    header = property(_get_header, doc="returns header only")

    def _get_dbtype(self):
        dbtype = self.header.split("|")[0].replace(">", "")
        return dbtype

    dbtype = property(_get_dbtype)

    # for all types
    def _get_identifier(self):
        return self.header.split(" ")[0]

    identifier = property(_get_identifier)

    def _get_entry(self):
        return self.header.split("|")[2].split(" ")[0]

    entry = property(_get_entry, doc="returns entry only")

    # swiss prot only
    def _get_accession(self):
        if self.dbtype == "sp":
            # header = self.header
            return self.identifier.split("|")[1]
        elif self.dbtype == "gi":
            return self.identifier.split("|")[1]

    accession = property(_get_accession)

    # swiss prot only
    def _get_name_sp(self):
        if self.dbtype == "sp":
            header = self.header
            return header.split(" ")[0].split("|")[2]

    name = property(_get_name_sp)

    @ifpandas
    def _get_df(self):
        import pandas as pd

        df = pd.DataFrame(
            {
                "Identifiers": [self.identifier],
                "Accession": [self.accession],
                "Entry": [self.entry],
                "Database": [self.dbtype],
                "Organism": [self.organism],
                "PE": [self.PE],
                "SV": [self.SV],
                "Sequence": [self.sequence],
                "Header": [self.header],
                "Size": [len(self.sequence)],
            }
        )
        return df

    df = property(_get_df)

    def _get_info_from_header(self, prefix):
        if prefix not in self.header:
            return None
        # finds the prefix
        index = self.header.index(prefix + "=")
        # remove it
        name = self.header[index:][3:]
        # figure out if there is anothe = sign to split the string
        # otherwise, the prefix we looked for is the last one anyway
        if "=" in name:
            name = name.split("=")[0]
            # here each = sign in FASTA is preceded by 2 characters that we must remove
            name = name[0:-2]
            name = name.strip()
        else:
            name = name.strip()
        return name

    def _get_gene_name(self):
        return self._get_info_from_header("GN")

    gene_name = property(
        _get_gene_name,
        doc="returns gene name from GN keyword found in the header if any",
    )

    def _get_organism(self):
        return self._get_info_from_header("OS")

    organism = property(_get_organism, doc="returns organism from OS keyword found in the header if any")

    def _get_PE(self):
        pe = self._get_info_from_header("PE")
        if pe is not None:
            return int(pe)

    PE = property(_get_PE, doc="returns PE keyword found in the header if any")

    def _get_SV(self):
        sv = self._get_info_from_header("SV")
        if sv is not None:
            return int(sv)

    SV = property(_get_SV, doc="returns SV keyword found in the header if any")

    def __str__(self):
        str_ = self.fasta
        return str_

    def get_fasta(self, id_):
        """Fetches FASTA from uniprot and loads into attrbiute :attr:`fasta`

        :param str id_: a given uniprot identifier
        :returns: the FASTA contents

        """
        print("get_fasta is deprecated. Use load_fasta instead")
        from bioservices import UniProt

        u = UniProt(verbose=False)
        res = u.retrieve(id_, frmt="fasta")
        self._fasta = res[:]
        return res

    def load(self, id_):
        self.load_fasta(id_)

    def load_fasta(self, id_):
        """Fetches FASTA from uniprot and loads into attribute :attr:`fasta`

        :param str id_: a given uniprot identifier
        :returns: nothing

        .. note:: same as :meth:`get_fasta` but returns nothing
        """
        # save fasta into attributes fasta
        from bioservices import UniProt

        u = UniProt(verbose=False)
        try:
            res = u.retrieve(id_, frmt="fasta")
            # some entries in uniprot are valid but obsolet and return empty string
            if res == "":
                raise Exception
            self._fasta = res[:]
        except:
            pass

    def save_fasta(self, filename):
        """Save FASTA file into a filename

        :param str data: the FASTA contents
        :param str filename: where to save it
        """
        if self._fasta == None:
            raise ValueError("No fasta was read or downloaded. Nothing to save.")

        fh = open(filename, "w")
        fh.write(self._fasta)
        fh.close()

    def read_fasta(self, filename):
        """Reads a FASTA file and loads it

        Type::

            >>> f = FASTA()
            >>> f.read_fasta(filename)
            >>> f.fasta

        :return: nothing

        .. warning:: If more than one FASTA is contained in the file, an error is raised
        """
        fh = open(filename, "r")
        data = fh.read()
        fh.close()

        # Is there more than one sequence ?
        data = data.split(">")[1:]
        if len(data) > 1 or len(data) == 0:
            raise ValueError(
                """Only one sequence expected to be found. Found %s. Please use MultiFASTA class instead""" % len(data)
            )

        self._data = data
        if data.count(">sp|") > 1:
            raise ValueError(
                """It looks like your FASTA file contains more than
            one FASTA. You must use MultiFASTA class instead"""
            )
        self._fasta = data[:]
        self._fasta = self._fasta[0]
        if self.dbtype not in self.known_dbtypes:
            print("Only sp and gi header are recognised so far but sequence and header are loaded")

    def _interpret(self, data):
        # cleanup the data in case of empty spaces or \n characters
        return data



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
# $Id$
"""Interface to some part of the Panther web service

.. topic:: What is Panther ?

    :URL: http://www.panther.org
    :Citation:

    .. highlights::


        The PANTHER (Protein ANalysis THrough Evolutionary Relationships)
        Classification System was designed to classify proteins (and
        their genes) in order to facilitate high-throughput analysis.
        Proteins have been classified according to:

            * Family and subfamily: families are groups of evolutionarily related
              proteins; subfamilies are related proteins that also have the same function
            * Molecular function: the function of the protein by itself or with directly
              interacting proteins at a biochemical level, e.g. a protein kinase
            * Biological process: the function of the protein in the context of a larger
              network of proteins that interact to accomplish a process at the level of the
              cell or organism, e.g. mitosis.
            * Pathway: similar to biological process, but a pathway also explicitly
              specifies the relationships between the interacting molecules.

        -- From PantherDB (about) , Feb 2020

"""
from bioservices.services import REST
from bioservices import logger

logger.name = __name__


__all__ = ["Panther"]


class Panther:
    """Interface to `Panther <http://www.pantherdb.org/services/oai/pantherdb>`_ pages


    ::

        >>> from bioservics import Panther
        >>> p = Panther()
        >>> p.get_supported_genomes()
        >>> p.get_ortholog("zap70", 9606)


        >>> from bioservics import Panther
        >>> p = Panther()
        >>> taxon = [x[0]['taxon_id'] for x in p.get_supported_genomes() if "coli" in x['name'].lower()]
        >>> # you may also use our method called search_organism
        >>> taxon = p.get_taxon_id(pattern="coli")
        >>> res = p.get_mapping("abrB,ackA,acuI", taxon)

    The get_mapping returns for each gene ID the GO terms corresponding to each
    ID. Those go terms may belong to different categories (see
    meth:`get_annotation_datasets`):

    - MF for molecular function
    - BP for biological process
    - PC for Protein class
    - CC Cellular location
    - Pathway

    Note that results from the website application http://pantherdb.org/
    do not agree with the oupput of the get_mapping service... Try out the dgt
    gene from ecoli for example




    """

    _url = "http://www.pantherdb.org/services/oai/pantherdb"

    def __init__(self, verbose=True, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        """
        # super(Panther, self).__init__(name="Panther", url=Panther._url,
        #       verbose=verbose, cache=cache)
        self.services = REST(
            name="Panther",
            url=Panther._url,
            verbose=verbose,
            cache=cache,
            url_defined_later=True,
        )

        self._allPathwaysURL = "http://www.pantherdb.org/pathway/pathwayList.jsp"

    def get_pathways(self):
        """Returns all pathways from pantherdb"""
        return self.services.http_get("supportedpantherpathways")

    def get_supported_genomes(self, type=None):
        """Returns list of supported organisms.

        :param type: can be chrLoc to restrict the search


        """
        if type is not None:
            params = {"type": type}
        else:
            params = {}
        res = self.services.http_get("supportedgenomes", params=params)
        res = [x for x in res["search"]["output"]["genomes"]["genome"]]
        return res

    def get_taxon_id(self, pattern=None):
        """return all taxons supported by the service

        If pattern is provided, we filter the name to keep those that contain
        the filter. If only one is found, we return the name itself, otherwise a
        list of candidates

        """
        res = self.get_supported_genomes()
        if pattern:
            taxon = [x["taxon_id"] for x in res if pattern.lower() in x["name"].lower()]
            if len(taxon) == 1:
                return taxon[0]
            else:
                return taxon
        else:
            taxon = [x["taxon_id"] for x in res]
            return taxon

    def get_mapping(self, gene_list, taxon):
        """Map identifiers

        Each identifier to be delimited by comma i.e. ',. Maximum of 1000 Identifiers
        can be any of the following: Ensemble gene identifier, Ensemble protein
        identifier, Ensemble transcript identifier, Entrez gene id, gene symbol, NCBI
        GI, HGNC Id, International protein index id, NCBI UniGene id, UniProt accession
        and UniProt id

        :param gene_list: see above
        :param taxon: one taxon ID. See supported
            :meth:`~bioservices.panther.Panther.get_supported_genomes`

        If an identifier is not found, information can be found in the
        unmapped_genes key while found identifiers are in the mapped_genes key.

        .. warning:: found and not found identifiers are dispatched into
            unmapped and mapped genes. If there are not found identifiers,
            the input gene list and the mapped genes list do not have the same
            length. The input names are not stored in the output.
            Developpers should be aware of that feature.

        """
        params = {"geneInputList": gene_list, "organism": taxon}
        res = self.services.http_post("geneinfo", params=params, frmt="json")

        if "mapped_genes" in res["search"]:
            mapped_genes = res["search"]["mapped_genes"]["gene"]
            # if only one identifier, retuns a dictionary.
            # if several identifiers, returns a list of dictionary.
            # We will be consistent and return a list
            if "accession" in mapped_genes:
                mapped_genes = [mapped_genes]
        else:
            mapped_genes = [{}]

        if "unmapped_list" in res["search"]:
            unmapped_genes = res["search"]["unmapped_list"]["unmapped"]
            if isinstance(unmapped_genes, list):
                pass
            else:
                unmapped_genes = [unmapped_genes]
        else:
            unmapped_genes = []

        logger.warning("Some identifiers were not found")
        return {"unmapped": unmapped_genes, "mapped": mapped_genes}

    def get_enrichment(
        self,
        gene_list,
        organism,
        annotation,
        enrichment_test="Fisher",
        correction="FDR",
        ref_gene_list=None,
    ):
        """Returns over represented genes

        Compares a test gene list to a reference gene list,
        and determines whether a particular class (e.g. molecular function,
        biological process, cellular component, PANTHER protein class, the
        PANTHER pathway or Reactome pathway) of genes is overrepresented
        or underrepresented.

        :param organism: a valid taxon ID
        :param enrichment_test: either **Fisher** or **Binomial** test
        :param correction: correction for multiple testing. Either **FDR**,
            **Bonferonni**, or **None**.
        :param annotation: one of the supported PANTHER annotation data types.
            See :meth:`~bioservices.panther.Panther.get_annotation_datasets` to retrieve a list of
            supported annotation data types
        :param ref_gene_list: if not specified, the system will use all the genes
            for the specified organism. Otherwise, a list delimited by
            comma. Maximum of 100000 Identifiers can be any of the
            following: Ensemble gene identifier, Ensemble protein
            identifier, Ensemble transcript identifier, Entrez gene id,
            gene symbol, NCBI GI, HGNC Id, International protein index id,
            NCBI UniGene id, UniProt accession andUniProt id.

        :return: a dictionary with the following keys. 'reference' contains the
            orgnaism, 'input_list' is the input gene list with unmapped genes.
            'result' contains the list of candidates.

        ::

            >>> from bioservices import Panther
            >>> p = Panther()
            >>> res = p.get_enrichment('zap70,mek1,erk', 9606, "GO:0008150")
            >>> For molecular function, use :
            >>> res = p.get_enrichment('zap70,mek1,erk', 9606,
                    "ANNOT_TYPE_ID_PANTHER_GO_SLIM_MF")

        """
        assert enrichment_test.lower() in ["fisher", "binomial"]
        if correction is None:
            correction = "none"

        assert correction.lower() in ["fdr", "bonferroni", "none"]

        # This is a bug in panther DB where they used bonferonni . should be
        # bonferroni...
        if correction.lower() == "bonferroni":
            correction = "bonferonni"
        assert annotation in [x["id"] for x in self.get_annotation_datasets()]

        params = {"enrichmentTestType": enrichment_test.upper()}
        params["organism"] = organism
        if gene_list:
            params["geneInputList"] = gene_list
        if ref_gene_list:
            params["refInputList"] = ref_gene_list
        params["annotDataSet"] = annotation
        params["correction"] = correction.upper()
        try:
            res = self.services.http_post("enrich/overrep", params=params, frmt="json")
            try:
                return res["results"]
            except:
                return res
        except:
            return res

    def get_annotation_datasets(self):
        """Retrieve the list of supported annotation data sets"""
        res = self.services.http_get("supportedannotdatasets")
        res = res["search"]["annotation_data_sets"]["annotation_data_type"]
        return res

    def get_ortholog(self, gene_list, organism, target_organism=None, ortholog_type="all"):
        """search for matching orthologs in target organisms.

        Searches for matching orthologs in the gene family that contains
        the search gene associated with the search terms. Returns
        ortholog genes in target organisms given a search organism,
        the search terms and a list of target organisms.

        :param gene_list:
        :param organism: a valid taxon ID
        :param target_organism: zero or more taxon IDs separated by ','. See
            :meth:`~bioservices.panther.Panther.get_supported_genomes`
        :param ortholog_type: optional parameter to specify ortholog type of target organism
        :return: a dictionary with "mapped" and "unmapped" keys, each of them
            being a list. For each unmapped gene, a dictionary with id and
            organism is is returned. For the mapped gene, a list of ortholog is
            returned.

        """
        assert ortholog_type in ["LDO", "all"]
        params = {
            "geneInputList": gene_list,
            "organism": organism,
            "targetOrganism": target_organism,
            "orthologType": ortholog_type,
        }
        if params["targetOrganism"] is None:
            del params["targetOrganism"]
        res = self.services.http_get("ortholog/matchortho", frmt="json", params=params)
        res = res["search"]["mapping"]
        mapped = res["mapped"]

        try:
            unmapped = res["unmapped_ids"]["unmapped"]
            # make sure we always have a list
            if isinstance(unmapped, dict):
                unmapped = [unmapped]
        except:
            unmapped = []
        res = {"unmapped": unmapped, "mapped": mapped}

        return res

    def get_homolog_position(self, gene, organism, position, ortholog_type="all"):
        """

        :param gene: Can be any of the following: Ensemble gene identifier,
            Ensemble protein identifier, Ensemble transcript identifier, Entrez gene id,
            gene symbol, NCBI GI, HGNC Id, International protein index id, NCBI UniGene id,
            UniProt accession andUniProt id
        :param organism: a valid taxon ID
        :param ortholog_type: optional parameter to specify ortholog type of target organism
        """
        if "," in gene:
            logger.warning("did not expect a comma. Please provide only one gene name")
        assert ortholog_type in ["LDO", "all"]
        assert position >= 1
        params = {
            "gene": gene,
            "organism": organism,
            "pos": position,
            "orthologType": ortholog_type,
        }
        res = self.services.http_get("ortholog/homologpos", params=params, frmt="json")
        res = res["search"]["mapping"]
        if "mapped" in res.keys():
            res = res["mapped"]
            return res
        elif "unmapped_ids" in res.keys():
            logger.warning("did not find any match for {}".format(gene))
            return res["unmapped_ids"]

    def get_supported_families(self, N=1000, progress=True):
        """Returns the list of supported PANTHER family IDs

        This services returns only 1000 items per request. This is defined by
        the index. For instance index set to 1 returns the first 1000 families.
        Index set to 2 returns families between index 1000 and 2000 and so on.
        As of 20 Feb 2020, there was about 15,000 families.

        This function simplifies your life by calling the service as many times
        as required. Therefore it returns all families in one go.

        """
        from easydev import Progress

        params = {"startIndex": 1}
        res = self.services.http_get("supportedpantherfamilies", params=params)
        results = res["search"]["panther_family_subfam_list"]["family"]
        if len(results) != N:
            msg = "looks like the services changed. Call this function with N={}"
            msg = msg.format(len(results))
            raise ValueError(msg)

        number_of_families = res["search"]["number_of_families"]
        pb = Progress(int(number_of_families / N))
        pb.animate(1)
        for i in range(1, int(number_of_families / N) + 1):
            params = {"startIndex": i * N + 1}
            res = self.services.http_get("supportedpantherfamilies", params=params)
            data = res["search"]["panther_family_subfam_list"]["family"]
            results.extend(data)
            if progress:
                pb.animate(i)
        return results

    def get_family_ortholog(self, family, taxon_list=None):
        """Search for matching orthologs in target organisms

        Also return the corresponding position in the target
        organism sequence. The system searches for matching
        orthologs in the gene family that contains the search
        gene associated with the search term.

        :param family: Family ID
        :param taxon_list: Zero or more taxon IDs separated by ','.
        """

        params = {"family": family}
        if taxon_list:
            params["taxonFltr"] = taxon_list
        res = self.services.http_get("familyortholog", params=params, frmt="json")
        return res["search"]["ortholog_list"]["ortholog"]

    def get_family_msa(self, family, taxon_list=None):
        """Returns MSA information for the specified family.

        :param family: family ID
        :param taxon_list: Zero or more taxon IDs separated by ','.

        """
        params = {"family": family}
        if taxon_list:
            params["taxonFltr"] = taxon_list
        res = self.services.http_get("familymsa", params=params, frmt="json")
        return res["search"]["MSA_list"]["sequence_info"]

    def get_tree_info(self, family, taxon_list=None):
        """Returns tree topology information and node attributes for the specified family.

        :param family: Family ID
        :param taxon_list: Zero or more taxon IDs separated by ','.
        """
        params = {"family": family}
        if taxon_list:
            params["taxonFltr"] = taxon_list
        res = self.services.http_get("treeinfo", params=params, frmt="json")
        return res["search"]  # ['tree_topology']['annotation_node']



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#  Copyright (c) 2022 - Institut Pasteur
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""This module provides a class :class:`UniChem`

.. topic:: What is UniChem

    :URL:  https://www.ebi.ac.uk/unichem/info/webservices
    :REST:  https://www.ebi.ac.uk/unichem/rest


    .. highlights::

        "UniChem is a 'Unified Chemical Identifier' system, designed to assist
        in the rapid cross-referencing of chemical structures, and their identifiers,
        between databases (read more). "

        -- From UniChem web page June 2013


"""
import json

from bioservices import REST
from bioservices import logger

import colorlog

logger = colorlog.getLogger(__name__)


class UniChem:
    """Interface to the `UniChem <https://www.ebi.ac.uk/unichem/>`_ service

    .. doctest::

            >>> from bioservices import UniChem
            >>> u = UniChem()

    There are lots of sources such as Chembl, Chebi, etc. You will probably need the
    identifiers of those sources. You can get all information about a source using
    these methods::

        # Get information about a source
        u.get_source_info_by_name('chembl')
        u.get_source_info_by_id(10)
        u.get_id_from_name('chembl')
        u.get_all_src_ids()

    but for developers, everything is contained in the :attr:`source_ids` dictionary.

    The first important method provided by Unichem API is the :meth:`get_compounds`.
    For example, you can request all compounds related to the CHEMBL12 identifier
    from ChEMBL using::

        res = u.get_compounds('CHEMBL12', 'chembl')
        compounds = res['compounds'][0]

    Note that the second argument is 'chembl' and lower/upper cases is important.
    All names are stored in :attr:`source_ids` together with their identifiers.

    You can use also :meth:`get_id_from_name` and get_name_from_id` if needed.

    Legacy methods are available:


        get_compound_ids_from_src_id            --> use get_compounds()
        get_src_compound_ids_from_inchikey      --> replaced by get_compounds()
        get_all_src_ids()                       --> uses new API
        get_src_compound_ids_all_from_inchikey  --> get_source_by_inchikey()
        get_verbose_src_compound_ids_from_inchikey  --> get_sources_by_inchikey_verbose()
        get_structure                           --> uses new API get_compounds() and bioservices code
        get_structure_all                       --> dropped
        get_src_compound_id_url                 --> dropped. One can use the get_compounds()
        get_src_compound_ids_all_from_obsolete  --> removed

        get_src_compound_ids_from_src_compound_id  --> removed; was obsolet
        get_src_compound_ids_all_from_src_compound_id --> remoed was already obsolet
        get_all_compound_ids_from_all_src_id   --> removed. no more API
        get_mapping                            --> removed. no more API
        get_auxiliary_mappings                 --> removed. no more API

    Most old functions can be replaced by a syntax such as::

        res = u.get_compound('CHEMBL12', 'chembl')
        res['compounds'][0]


    .. changed:: version 1.9. drop xml parser.


    """

    _url = "https://www.ebi.ac.uk/unichem"

    def __init__(self, verbose=False, cache=False):
        """.. rubric:: **Constructor** UniChem

        :param verbose: set to False to prevent informative messages
        """
        self.services = REST(name="UniChem", url=UniChem._url, verbose=verbose, cache=cache)

        # let us define the source and names
        _data = self.services.http_get("api/v1/sources")
        self._data_source = _data["sources"]
        self.source_ids = {x["name"]: x["sourceID"] for x in self._data_source}

    def get_id_from_name(self, name):
        """Return the ID a a source given its name.

        :param str name: a valid database name (e.g., chembl)

        ::

            u.get_id_from_name("chembl")
        """
        if name in self.source_ids.keys():
            return self.source_ids[name]
        else:
            logger.error(f"You provided {name} but only those sources are available: {sorted(self.source_ids.keys())}")

    def get_sources(self):
        """Returns all information about all sources used in Unichem
        ::

            from bioservices import UniChem
            u = UniChem()
            res = u.get_sources_information()
            res['sources']
        """
        return self._data_source

    # NEWS
    def get_inchi_from_inchikey(self, inchikey):
        """Get a list of inchis given a valid inchikey.

        :param inchikey: InChI Key to search. Unlike the rest API, you can also provide a list.
        :return: a list of inchis matching the InChI Key provided. If input is a list, a
            dictionary is returned where keys are the inchikey input lists.

        ::

            from bioservices import UniChem
            u = UniChem()
            res = u.get_inchi("AAOVKJBEBIDNHE-UHFFFAOYSA-N")

        .. note:: this is a legacy function. introduced in v1.9 after unichem API update
        """
        # if inchikey is not found, return empty {}
        if isinstance(inchikey, (list, tuple)):
            data = {}
            for x in inchikey:
                res = self.services.http_get(f"rest/inchi/{x}")
                data[x] = {} if res == 500 else res
            return data
        else:
            res = self.services.http_get(f"rest/inchi/{inchikey}")
            res = {} if res == 500 else res
            return res

    def get_sources_by_inchikey(self, inchikey):
        """Get sources by inchikey

        :param inchikey: InChI Key to search. Unlike the rest API, you can also provide a list.
        :return: A list of sources for the provided InChIKey if input is a single string.
            a dictionary with keys as inchikey if input is a list.


        .. note:: this is a legacy function. introduced in v1.9 after unichem API update
        """
        # if inchikey is not found, return empty {}
        if isinstance(inchikey, (list, tuple)):
            data = {}
            for x in inchikey:
                res = self.services.http_get(f"rest/inchikey/{x}")
                data[x] = {} if res == 500 else res
            return data
        else:
            res = self.services.http_get(f"rest/inchikey/{inchikey}")
            res = {} if res == 500 else res
            return res

    def get_sources_by_inchikey_verbose(self, inchikey):
        """Get sources by inchikey

        :param inchikey: InChI Key to search. Unlike the rest API, you can also provide a list.
        :return: A list of sources for the provided InChIKey if input is a single string.
            a dictionary with keys as inchikey if input is a list.


        .. note:: this is a legacy function. introduced in v1.9 after unichem API update
        """
        # if inchikey is not found, return empty {}
        if isinstance(inchikey, (list, tuple)):
            data = {}
            for x in inchikey:
                res = self.services.http_get(f"rest/verbose_inchikey/{x}")
                data[x] = {} if res == 500 else res
            return data
        else:
            res = self.services.http_get(f"rest/verbose_inchikey/{inchikey}")
            res = {} if res == 500 else res
            return res

    def get_all_src_ids(self):
        """Obtain all src_ids of sources available in UniChem

        :return: list of 'src_id's.

        ::

            uni.get_all_src_ids()

        """
        return sorted([x["sourceID"] for x in self._data_source])

    def get_source_info_by_name(self, src_name):
        """Description:  Obtain all information on a source by querying with a source id

        :param int src_name: valid identifiers can be found in :attr:`source_ids` e.g.
            chebi, chembl)
        :return: dictionary (or list of dictionaries) with following keys:

            * UCICount: number of entries
            * baseIdUrl: URL of the source
            * created: date of creation
            * description: a description of the content of the source
            * lastUpdated: last date of the update
            * name: the unique name for the source in UniChem, always lower case
            * nameLabel: A name for the source suitable for use as a 'label' for the source
            * nameLong: the full name of the source, as defined by the source
            * private: is it private or not ?
            * sourceID: the src_id for this source
            * srcDetails: details about the source
            * srcReleaseDate: release date of the source database
            * srcReleaseNumber: release number of the source
            * srcUrl: src_url (the main home page of the source)
            * updateComments: possible updates from this source

        ::

            >>> res = get_source_by_name("chebi")

        """
        keys = sorted([x["name"] for x in self._data_source])
        if src_name in keys:
            return [x for x in self._data_source if x["name"] == src_name][0]

        logger.warning(f"incorrect {src_name} source name. Use one of {keys}")

    def get_source_info_by_id(self, ID):
        ids = sorted([x["sourceID"] for x in self._data_source])

        if ID in ids:
            return [x for x in self._data_source if x["sourceID"] == ID][0]

        logger.warning(f"incorrect {ID} source name. Use one of {ids}")

    def get_compounds(self, compound, source_type):
        """Get matched compounds information

        :param str compound: InChI, InChIKey, Name, UCI or Compound Source ID
        :param source_type: uci, inchi, inchikey, sourceID (e.g. chembl)
        :param str sourceID: ID for the source assigned in UniChem when the type is "sourceID"
        :return: a list of matched compounds and their assigned sources


        A legacy function allows you to retrieve a compound from its inchikey::

            u.get_sources_by_inchikey('GZUITABIAKMVPG-UHFFFAOYSA-N')

        However, this new function is faster presumably and allows you to do the same::

            res = u.get_compounds('GZUITABIAKMVPG-UHFFFAOYSA-N', 'inchikey')
            res['compounds']

        You can get the first element, from which inchi, sources, standardInchikey, uci can be extracted.
        The **sources** key contains all compound identifiers for each source::

            res['compounds'][0]['uci']
            res['compounds'][0]['sources']

        Looks like there is always a single element in res['compounds'] but since it is a list,
        you must access to first element (unique) using [0] syntax.

        """
        # we need a default value set to empty string
        sourceID = ""

        # source type can be either one of:
        if source_type in ["uci", "inchi", "inchikey"]:
            pass
        # or a valid source identifier from a valid source name.
        elif source_type in self.source_ids.keys():
            sourceID = int(self.source_ids[source_type])
            source_type = "sourceID"
        # or simply the valid source identifier
        elif source_type in self.source_ids.values():
            sourceID = source_type
            source_type = "sourceID"
        else:
            logger.error(
                f"source_type must be one of uci, inchi, inchikey or a valid source from {sorted(self.source_ids.keys())}"
            )
            return {}

        body = {"compound": compound, "sourceID": sourceID, "type": source_type}
        # somehow, the expected input is a json string and output a json string but cannot be
        # encode/devode by the request even though we provide fmrt=json
        body = json.dumps(body)
        res = self.services.http_post("api/v1/compounds", data=body, headers=self.services.get_headers("json"))
        try:  # pragma: no cover
            res = json.loads(res)
            return res
        except TypeError:  # pragma: no cover
            return {}

    def get_connectivity(self, compound, source_type):
        """Fetch multiple source data sets for a given compound
        with common connectivity to a given id on the database
        source, InChI, InChIkey or UCI

        :param str compound: InChI, InChIKey, Name, UCI or Compound Source ID (e.g. chembl)
        :param source_type: uci, inchi, inchikey, sourceID

        The returned dictionary contains 5 keys:

        * response: service response ('Success' if everything is right)
        * searchedCompound: the summary in terms of inchi, standardInchikey and uci
        * sources: a dictionary with e.g. compoundID and name of the source.
            A 'comparison' dictionary is also provided.
        * totalCompounds: number of searchedCompound entries
        * totalSources: number of sources entries

        """

        # we need a default value set to empty string
        sourceID = ""

        # source type can be either one of:
        if source_type in ["uci", "inchi", "inchikey"]:
            pass
        # or a valid source identifier from a valid source name.
        elif source_type in self.source_ids.keys():
            sourceID = int(self.source_ids[source_type])
            source_type = "sourceID"
        # or simply the valid source identifier
        elif source_type in self.source_ids.values():
            sourceID = source_type
            source_type = "sourceID"
        else:
            logger.error(
                f"source_type must be one of uci, inchi, inchikey or a valid source from {sorted(self.source_ids.keys())}"
            )
            return {}

        body = {"compound": compound, "sourceID": sourceID, "type": source_type}
        # somehow, the expected input is a json string and output a json string but cannot be
        # encode/devode by the request even though we provide fmrt=json
        body = json.dumps(body)
        res = self.services.http_post("api/v1/connectivity", data=body, headers=self.services.get_headers("json"))
        try:  # pragma: no cover
            res = json.loads(res)
            return res
        except TypeError:  # pragma: no cover
            return {}

    def get_images(self, uci, filename=None):
        """Return / create compound image

        :param uci: the UCI of the compound
        :param filename: optional file name to save the SVG+XML output
        :return: the SVG+XML string

        .. plot::

            res = u.get_images('304698', filename='test.svg')

        """
        res = self.services.http_get(f"api/v1/images/{uci}", headers=self.services.get_headers("svg+xml"))
        try:
            res = res.content
            if filename:
                with open(filename, "w") as fout:
                    fout.write(res.decode())
            return res
        except AttributeError:
            logger.warning("Invalid UCI request")

    # OLD ------------------------------
    def get_structure(self, compound_id, src_id):
        """Obtain structure(s) CURRENTLY assigned to a query src_compound_id.

        :param str compound_id: a valid compound identifier
        :param int src_id: corresponding database identifier (name or id).

        :return:  dictionary with 'standardinchi' and 'standardinchikey' keys

        ::

            >>> uni.get_structure("CHEMBL12", "chembl")

        """

        res = self.get_compounds(compound_id, src_id)
        res = res["compounds"]
        res = res[0]
        return {"inchi": res["inchi"]["inchi"], "standardInchiKey": res["standardInchiKey"]}



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to some part of the EVA web service

.. topic:: What is EVA ?

    :URL: http://www.uniprot.org
    :Citation:

    .. highlights::

        "The European Variation Archive is an open-access database of all
        types of genetic variation data from all species. The EVA provides
        access to highly detailed, granular, raw variant data from human, with
        other species to follow."

        -- From Uniprot web site (help/about) , Dec 2012


"""
import types
import io

from bioservices.services import REST
from bioservices import logger

logger.name = __name__


__all__ = ["EVA"]


class EVA(REST):
    """Interface to the `EVA <http://www.ebi.ac.uk/eva>`_ service


    * version: indicates the version of the API, this defines the available
      filters and JSON schema to be returned. Currently there is only
      version 'v1'.
    * category: this defines what objects we want to query. Currently there
      are five different categories: variants, segments, genes, files and
      studies.
    * resource: specifies the resource to be returned, therefore the JSON
      data model.
    * filters: each specific endpoint allows different filters.
    """

    _url = "http://wwwdev.ebi.ac.uk/eva/webservices/rest/"

    def __init__(self, verbose=False, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        """
        super(EVA, self).__init__(name="EVA", url=EVA._url, verbose=verbose, cache=cache)
        self.version = "v1"

    def fetch_allinfo(self, name):
        """e.g., PRJEB4019"""
        res = self.http_get(self.version + "/studies/" + name + "/summary")
        return res



from bioservices import UniProt
import re


class Peptides(object):
    """

    ::

        >>> p = Peptides()
        >>> p.get_peptide_position("Q8IYB3", "VPKPEPIPEPKEPSPE")
        189


    Sometimes, peptides are provided with a pattern indicating the phospho site.
    e.g., ::

        >>>

    """

    def __init__(self, verbose=False):
        self.u = UniProt(verbose=verbose)
        self.sequences = {}

    def get_fasta_sequence(self, uniprot_name):
        seq = self.u.get_fasta(uniprot_name)
        seq = "".join(seq.split("\n")[1:])
        return seq


    def get_phosphosite_position(self, uniprot_name, peptide):
        if uniprot_name not in self.sequences.keys():
            seq = self.get_fasta_sequence(uniprot_name)
            self.sequences[uniprot_name] = seq[:]
        else:
            seq = self.sequences[uniprot_name][:]
        positions = [x.start() for x in re.finditer("PQS", seq)]
        return positions



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to Ensembl web service

.. topic:: What is Ensembl ?

    :URL: http://www.ensembl.org/
    :URL: https://github.com/Ensembl/ensembl-rest/wiki/Output-formats

    .. highlights::

        The Ensembl project produces genome databases for vertebrates and
        other eukaryotic species, and makes this information freely available
        online.

        -- From Ensembl web site, Oct 2014


"""
from bioservices.services import REST
from bioservices import logger

logger.name = __name__


__all__ = ["Ensembl"]


class Ensembl:
    """Interface to the `Ensembl <http://rest.ensembl.org>`_ service

    For the BioServices documentation see the documentation of
    each method for the list of parameters. The API was copied
    from the Ensemble API (http://rest.ensembl.org)

    All methods have been tests using this BioServices
    `notebook <http://nbviewer.ipython.org/github/bioservices/notebooks/blob/master/ensembl/Ensembl.ipynb>`_


    .. todo:: There are 3 methods out of 50 that are not implemented so far.
    .. todo:: some methods have a parameter called *feature*. The official
       Ensembl API allows one to provide several features at the same time.
       This is not yet implemented. Only one at a time is accepted.

    .. note:: Some function uses SQL wildcards. See e.g. http://www.w3schools.com/sql/sql_wildcards.asp
        In brief, "_" can be use to substitute a single character and '%' a set of characters.
    """

    _url = "http://rest.ensembl.org"

    def __init__(self, verbose=False, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        """
        self.services = REST(name="Ensembl", url=Ensembl._url, verbose=verbose, cache=cache)
        self.callback = None  # use in all methods

    def _check_frmt(self, frmt, values=[]):
        self.services.devtools.check_param_in_list(frmt, ["json", "jsonp"] + values)

    def _check_id(self, identifier):
        pass

    def nh_format_to_frmt(self, value):
        if value == "phylip":
            return "phyloxml"
        elif value == "simple":
            return "nh"
        else:
            return "json"

    def check_sequence(self, value):
        self.services.devtools.check_param_in_list(value, ["none", "cdna", "protein"])

    def check_nh_format(self, value):
        self.services.devtools.check_param_in_list(
            value,
            [
                "full",
                "display_label_composite",
                "simple",
                "species",
                "species_short_name",
                "ncbi_taxon",
                "ncbi_name",
                "njtree",
                "phylip",
            ],
        )

    # ARCHIVE
    # ---------------------------------------------------------------------
    def get_archive(self, identifier, frmt="json"):
        """Uses the given identifier to return the archived sequence

        :param str identifier: An Ensembl stable ID
        :param str frmt: output formart(json, xml or jsonp)

        ::

            >>> from bioservices import Ensembl
            >>> s = Ensembl()
            >>> res = s.get_archive("ENSG00000157764")

        """
        self._check_frmt(frmt, ["xml"])
        self._check_id(identifier)
        res = self.services.http_get(
            "archive/id/" + identifier,
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )

        if frmt == "xml":
            res = self.easyXML(res)
        return res

    def post_archive(self, identifiers, frmt="json"):
        """Retrieve the archived sequence for a set of identifiers

        returned by the requested JSONP response. Required ONLY when using
        JSONP as the serialisation method. Please see the user guide.

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_post(
            "archive/id/",
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            data=self.services.devtools.to_json({"id": identifiers}),
            params={"callback": self.callback},
        )
        if frmt == "xml":
            res = self.easyXML(res)
        return res

    # COMPARATIVE GENOMICS
    # ------------------------------------------------------------------

    def get_genetree_by_member_symbol(
        self,
        species,
        symbol,
        frmt="json",
        aligned=False,
        db_type="core",
        object_type=None,
        nh_format="simple",
        sequence="protein",
        compara="multi",
    ):
        """Retrieves a gene tree containing the gene identified by a symbol"""
        self._check_frmt(frmt, ["nh", "phyloxml"])
        frmt = self.nh_format_to_frmt(nh_format)
        self.check_sequence(sequence)
        self.check_nh_format(nh_format)

        res = self.services.http_get(
            "genetree/member/symbol/%s/%s" % (species, symbol),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "nh_format": nh_format,
                "sequence": sequence,
                "aligned": int(aligned),
                "compara": compara,
                "db_type": db_type,
            },
        )
        return res

    def get_genetree_by_member_id(
        self,
        identifier,
        frmt="json",
        aligned=False,
        db_type="core",
        object_type=None,
        nh_format="simple",
        sequence="protein",
        species="homo_sapiens",
        compara="multi",
    ):
        """Retrieves a gene tree containing the gene identified by a symbol

        :param str compara: Name of the compara database to use. Multiple
            comparas can exist on a server if you are accessing Ensembl
            Genomes data. Defautl to 'multi'
        :param db_type: Restrict the search to a database other than the
            default. Useful if you need to use a DB other than core. Defaults
            to core
        :param object_type: Filter by feature type. Default to None
            examples are gene, transcript.

        :

            get_genetree_by_member_id('ENSG00000157764', frmt='phyloxml')


        """
        self._check_frmt(frmt, ["nh", "phyloxml"])
        self._check_frmt(frmt)
        self.check_nh_format(nh_format)
        frmt = self.nh_format_to_frmt(nh_format)
        self.check_sequence(sequence)

        res = self.services.http_get(
            "genetree/member/id/" + identifier,
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "nh_format": nh_format,
                "sequence": sequence,
                "aligned": int(aligned),
                "compara": compara,
                "db_type": db_type,
                "species": species,
            },
        )
        return res

    def get_genetree_by_id(
        self,
        identifier,
        aligned=False,
        frmt="json",
        nh_format="simple",
        sequence="protein",
        compara="multi",
    ):
        """Retrieves a gene tree dump for a gene tree stable identifier

        :param str identifier: An Ensembl genetree ID
        :param str frmt: response formats: json, jsonp, nh, phyloxml
        :param bool aligned: if true, return the aligned string otherwise, i
            return the original
            sequence (no insertions). Can be True/1 or False/0 and defaults to 0
        :param str compara: Name of the compara database to use. Multiple comparas can
            exist on a server if you are accessing Ensembl Genomes data
        :param nh_format: The format of a NH (New Hampshire) request.
            Valid values are 'full', 'display_label_composite', 'simple',
            'species', 'species_short_name', 'ncbi_taxon', 'ncbi_name',
            'njtree', 'phylip'
        :param sequence: The type of sequence to bring back. Setting it to none
            results in no sequence being returned.
            Valid values are 'none', 'cdna', 'protein'.

        ::

            >>> from bioservices import Ensembl
            >>> s = Ensembl()
            >>> s.get_genetree('ENSGT00390000003602', frmt='nh', nh_format='simple')
            >>> s.get_genetree('ENSGT00390000003602', frmt='phyloxml')
            >>> s.get_genetree('ENSGT00390000003602', frmt='phyloxml',aligned=True, sequence='cdna')
            >>> s.get_genetree('ENSGT00390000003602', frmt='phyloxml', sequence='none')

        """
        self._check_frmt(frmt, ["nh", "phyloxml"])
        self.check_nh_format(nh_format)
        aligned = int(aligned)
        self.check_sequence(sequence)
        res = self.services.http_get(
            "genetree/id/" + identifier,
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"nh_format": nh_format, "sequence": sequence, "aligned": aligned},
        )
        return res

    def get_alignment_by_region(
        self,
        region,
        species,
        frmt="json",
        aligned=True,
        compact=True,
        compara="multi",
        display_species_set=None,
        mask=None,
        method="EPO",
        species_set=None,
        species_set_group="mammals",
    ):
        """Retrieves genomic alignments as separate blocks based on a region and
        species

        :param str region: Query region. A maximum of 10Mb is allowed to be
            requested at any one time (e.g.,  'X:1000000..1000100:1',
            'X:1000000..1000100:-1',  'X:1000000..1000100')
        :param str species: Species name/alias (e.g., human)
        :param bool aligned: Return the aligned string if true. Otherwise,
            return the original sequence (no insertions)
        :param bool compact: Applicable to EPO_LOW_COVERAGE alignments.
            If true, concatenate the low coverage species sequences
            together to create a single sequence. Otherwise, separates
            out all sequences.
        :param str compara: Name of the compara database to use. Multiple
            comparas can exist on a server if you are accessing Ensembl
            Genomes data (defaults to multi)
        :param str display_species_set: Subset of species in the alignment
            to be displayed (multiple values). All the species in the alignment
            will be displayed if this is not set. Any valid alias may be
            used.. (e.g., human, chimp, gorilla)
        :param str mask: Request the sequence masked for repeat sequences.
            Hard will mask all repeats as N's and soft will mask repeats
            as lowercased characters.
        :param str method:The alignment method amongst Enum(EPO,
            EPO_LOW_COVERAGE, PECAN, LASTZ_NET, BLASTZ_NET, TRANSLATED_BLAT_NET)
        :param str species_set: the set of species used to define the pairwise
            alignment (multiple values). Should not be used with the
            species_set_group parameter. Use :meth:`get_info_compara_by_method`
            with one of the methods listed above to obtain a valid list of
            species sets. Any valid alias may be used. (e.g., musc_musculus,
            homo_sapiens)
        :param str species_set_group: The species set group name of the multiple
            alignment. Should not be used with the species_set parameter.
            Use /info/compara/species_sets/:method with one of the methods
            listed above to obtain a valid list of group names. (Defaults to
            mammals. e.g. mammals, amniotes, fish, sauropsids)


        """
        self._check_frmt(frmt, ["xml", "phyloxml"])
        res = self.services.http_get(
            "alignment/region/{0}/{1}".format(species, region),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "aligned": int(aligned),
                "callback": self.callback,
                "compact": compact,
                "compara": compara,
                "display_species_set": display_species_set,
                "mask": mask,
                "method": method,
                "species_set": species_set,
                "species_set_group": species_set_group,
            },
        )
        return res

    def get_homology_by_id(
        self,
        identifier,
        frmt="json",
        aligned=True,
        compara="multi",
        format=None,
        sequence=None,
        species=None,
        target_species=None,
        target_taxon=None,
        type="all",
    ):
        """Retrieves homology information (orthologs) by Ensembl gene id"""
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "homology/id/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "aligned": int(aligned),
                "callback": self.callback,
                "compara": compara,
                "format": format,
                "sequence": sequence,
                "species": species,
                "target_species": target_species,
                "target_taxon": target_taxon,
                "type": type,
            },
        )
        return res

    def get_homology_by_symbol(self, species, symbol, frmt="json", aligned=True, compara=None):
        """Retrieves homology information (orthologs) by symbol"""
        raise NotImplementedError("Please fill an issue her https://github.com/cokelaer/bioservices")
        # species='human', symbol='BRCA2'

    # CROSS REFERENCES
    # -------------------------------

    def get_xrefs_by_id(
        self,
        identifier,
        frmt="json",
        all_levels=False,
        db_type="core",
        external_db=None,
        object_type=None,
        species=None,
    ):
        """Perform lookups of Ensembl Identifiers and retrieve their external
        references in other databases

        :param str identifier: An Ensembl Stable ID (ENSG00000157764)
        :param str frmt: response formats: json, jsonp, nh, phyloxml
        :param bool all_levels: Set to find all genetic features linked to
            the stable ID, and fetch all external references for them.
            Specifying this on a gene will also return values from its
            transcripts and translations.
        :param str db_type: Restrict the search to a database other than the
            default. Useful if you need to use a DB other than core
        :param str external_db: Filter by external database (e.g., HGNC)
        :param str object_type: filter by feature type (e.g., gene, transcript)
        :param str species: Species name/alias (human)

        """
        self._check_frmt(frmt)
        res = self.services.http_get(
            "xrefs/id/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "db_type": db_type,
                "callback": self.callback,
                "db_type": db_type,
                "all_levels": int(all_levels),
                "external_db": external_db,
                "object_type": object_type,
                "species": species,
            },
        )
        return res

    def get_xrefs_by_name(self, name, species, frmt="json", db_type="core", external_db=None):
        """Performs a lookup based upon the primary accession or display label
        of an external reference and returning the information we hold about the
        entry

        :param str name: Symbol or display name of a gene (e.g., BRCA2)
        :param str species: Species name/alias (e.g., human)
        :param str frmt: response formats: json, jsonp,xml
        :param str db_type: Restrict the search to a database other than the
            default. Useful if you need to use a DB other than core
        :param str external_db: Filter by external database (e.g., HGNC)


        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "xrefs/name/{0}/{1}".format(species, name),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "db_type": db_type,
                "callback": self.callback,
                "external_db": external_db,
            },
        )
        return res

    def get_xrefs_by_symbol(
        self,
        symbol,
        species,
        frmt="json",
        db_type="core",
        external_db=None,
        object_type=None,
    ):
        """Looks up an external symbol and returns all Ensembl objects linked to
        it. This can be a display name for a gene/transcript/translation, a
        synonym or an externally linked reference. If a gene's transcript is
        linked to the supplied symbol the service will return both gene and
        transcript (it supports transient links).

        :param str species: Species name/alias (e.g., human)
        :param str symbol: Symbol or display name of a gene (BRCA2)
        :param str frmt: response formats: json, jsonp,xml
        :param str db_type: Restrict the search to a database other than the
            default. Useful if you need to use a DB other than core
        :param str external_db: Filter by external database (e.g., HGNC)
        :param str object_type: filter by feature type (e.g., gene, transcript)

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "xrefs/symbol/{0}/{1}".format(species, symbol),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            #      params = {'db_type': db_type, 'callback':self.callback,
            #           'object_type':object_type, 'external_db': external_db}
        )
        return res

    # INFORMATION
    # ---------------------------------------------------------------------------
    def get_info_analysis(self, species, frmt="json"):
        """List the names of analyses involved in generating Ensembl data.

        :param str species: Species name/alias (e.g., homo_sapiens)
        :param str frmt: response formats: json, jsonp,xml

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/analysis/{0}".format(species),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    def get_info_assembly(self, species, frmt="json", bands=False):
        """List the currently available assemblies for a species.

        :param str species: Species name/alias (e.g., homo_sapiens)
        :param str frmt: response formats: json, jsonp,xml
        :param bool bands: if set to 1, include karyotype band information.
            Only display if band information is available

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/assembly/{0}".format(species),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"bands": bands, "callback": self.callback},
        )
        return res

    def get_info_assembly_by_region(self, species, region, frmt="json", bands=0):
        """Returns information about the specified toplevel sequence region for the
        given species."""
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/assembly/{0}/{1}".format(species, region),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"bands": bands, "callback": self.callback},
        )
        return res

    def get_info_biotypes(self, species, frmt="json"):
        """List the functional classifications of gene models that Ensembl associates
        with a particular species. Useful for restricting the type of genes/transcripts
        retrieved by other endpoints.

        :param str species: Species name/alias (e.g., homo_sapiens)
        :param str frmt: response formats: json, jsonp,xml

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/biotypes/{0}".format(species),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    def get_info_compara_methods(self, frmt="json", compara="multi", method_class=None):
        """List all compara analyses available (an analysis defines the type of
        comparative data).

        :param str frmt: response formats: json, yaml, jsonp, xml
        :param str class: The class of the method to query for.
            Regular expression patterns are supported.
            (Defaults to  GenomicAlign)
        :param str compara: Name of the compara database to use.
            Multiple comparas may exist on a server when accessing Ensembl Genomes data.

        .. note:: API argument is class, renamed in method_class

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/compara/methods",
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "class": method_class,
                "compara": compara,
                "callback": self.callback,
            },
        )
        return res

    def get_info_compara_by_method(
        self,
        method,
        frmt="json",
        compara="multi",
    ):
        """List all collections of species analysed with the specified compara
        method.

        :param str method: Filter by compara method. Use one the
            methods returned by /info/compara/methods endpoint.
            e.g., EPO
        :param str frmt: response formats: json, jsonp,xml
        :param str compara: Name of the compara database to use. Multiple
            comparas may exist on a server when accessing Ensembl Genomes data.
            defaults to 'multi'

        """
        self._check_frmt(frmt, ["xml", "yaml"])
        res = self.services.http_get(
            "info/compara/species_sets/{0}".format(method),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"compara": compara, "callback": self.callback},
        )
        return res

    def get_info_comparas(self, frmt="json"):
        """Lists all available comparative genomics databases and their data release.

        :param str frmt: response formats: json, jsonp,xml
        """
        self._check_frmt(frmt, ["xml", "yaml"])
        res = self.services.http_get(
            "info/comparas",
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    def get_info_data(self, frmt="json"):
        """Shows the data releases available on this REST server. May return more than
        one release (unfrequent non-standard Ensembl configuration).

        :param str frmt: response formats: json, jsonp,xml


        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/data",
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    def get_info_external_dbs(self, species, frmt="json", filter=None):
        """Lists all available external sources for a species.

        :param str frmt: response formats: json, jsonp,xml
        :param str species: Species name/alias
        :param str filter: Restrict external DB searches to a single
            source or pattern. SQL-LIKE patterns are supported.
            See :class:`Ensembl` doc.


        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/external_dbs/{0}".format(species),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback, "filter": filter},
        )
        return res

    def get_info_ping(self, frmt="json"):
        """Checks if the service is alive."""
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/ping/",
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res["ping"]

    def get_info_rest(self, frmt="json"):
        """Shows the current version of the Ensembl REST API.

        :param str frmt: response formats: json, jsonp,xml

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/rest/",
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    def get_info_software(self, frmt="json"):
        """Shows the current version of the Ensembl API used by the REST server.

        :param str frmt: response formats: json, jsonp,xml
        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/software/",
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    def get_info_species(self, frmt="json"):
        """Lists all available species, their aliases, available adaptor groups and data
        release.

        :param str frmt: response formats: json, jsonp,xml

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "info/species/",
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    # LOOKUP
    # -------------------------------------------------------------------
    def get_lookup_by_id(
        self,
        identifier,
        frmt="json",
        db_type=None,
        expand=False,
        format="full",
        species=None,
    ):
        """Find the species and database for a single identifier

        :param str identifier: An ontology term identifier (e.g., GO:0005667)
        :param str frmt: response formats in json, xml, jsonp
        :param str db_type: Restrict the search to a database other than the
            default. Useful if you need to use a DB other than core. Defaults
            to core
        :param str expand: Expands the search to include any connected features.
            e.g. If the object is a gene, its transcripts, translations and exons
            will be returned as well.
        :param str format: Specify the formats to emit from this endpoint
        :param str species: Species name/alias (e.g., human)

        ::

            get_lookup_by_id('ENSG00000157764', expand=True)

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "lookup/id/" + identifier,
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "db_type": db_type,
                "expand": int(expand),
                "format": format,
                "callback": self.callback,
                "species": species,
            },
        )
        return res

    def post_lookup_by_id(
        self,
        identifiers,
        frmt="json",
        db_type=None,
        expand=False,
        format="full",
        object_type=None,
        species=None,
    ):
        """Find the species and database for a single identifier

        :param str identifier: An ontology term identifier (e.g., GO:0005667)
        :param str frmt: response formats in json, xml, jsonp
        :param str db_type: Restrict the search to a database other than the
            default. Useful if you need to use a DB other than core. Defaults
            to core
        :param str expand: Expands the search to include any connected features.
            e.g. If the object is a gene, its transcripts, translations and exons
            will be returned as well.
        :param str format: Specify the formats to emit from this endpoint
        :param str object_type: Filter by feature type (e.g., gene, transcript)
        :param str species: Species name/alias (e.g., human)

        ::

            post_lookup_by_id(["ENSG00000157764", "ENSG00000248378" ])

        .. todo: frmt can only be json or jsonp
        """
        self._check_frmt(frmt)
        identifiers = self.services.devtools.to_list(identifiers)
        expand = int(expand)
        res = self.services.http_post(
            "lookup/id/",
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            data=self.services.devtools.to_json({"ids": identifiers}),
            params={
                "db_type": db_type,
                "expand": expand,
                "format": format,
                "callback": self.callback,
                "species": species,
            },
        )
        return res

    def get_lookup_by_symbol(
        self,
        species,
        symbol,
        frmt="json",
        expand=False,
        format="full",
    ):
        """Find the species and database for a single identifier

        :param str species: Species name/alias (e.g., human)
        :param str symbol: A name or symbol from an annotation source has been
            linked to a genetic feature. e.g., BRCA2
        :param str frmt: response formats in json, xml, jsonp
        :param str expand: Expands the search to include any connected features.
            e.g. If the object is a gene, its transcripts, translations and exons
            will be returned as well.
        :param str format: Specify the formats to emit from this endpoint

        ::

            get_lookup_by_symbol('homo_sapiens', 'BRCA2', expand=True)

        """
        self._check_frmt(frmt, ["xml"])
        expand = int(expand)
        res = self.services.http_get(
            "lookup/symbol/{0}/{1}".format(species, symbol),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"format": format, "callback": self.callback, "expand": expand},
        )
        return res

    def post_lookup_by_symbol(
        self,
        species,
        symbols,
        frmt="json",
        expand=False,
        format="full",
    ):
        """Find the species and database for a set of symbols

        :param str species: Species name/alias (e.g., human)
        :param list symbols: A list of names or symbols from an annotation source has been
            linked to a genetic feature. e.g., BRCA2
        :param str frmt: response formats in json, xml, jsonp
        :param str expand: Expands the search to include any connected features.
            e.g. If the object is a gene, its transcripts, translations and exons
            will be returned as well.
        :param str format: Specify the formats to emit from this endpoint

        ::

            post_lookup_by_symbol('homo_sapiens', ['BRCA2', 'BRAF'], expand=True)

        """
        self._check_frmt(frmt, ["xml"])
        symbols = self.services.devtools.to_list(symbols)
        expand = int(expand)
        res = self.services.http_post(
            "lookup/symbol/{0}".format(species),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            data=self.services.devtools.to_json({"symbols": symbols}),
            params={"format": format, "callback": self.callback, "expand": expand},
        )
        return res

    # MAPPING
    # --------------------------------------------------------------------

    def get_map_cds_to_region(self, identifier, region, frmt="json", species=None):
        """Convert from cDNA coordinates to genomic coordinates.

        :param identifier: Ensembl ID e.g. ENST00000288602
        :param region: Query region e.g., 100..300

        Output reflects forward orientation coordinates as returned from the Ensembl API.


            get_map_cds_to_region('ENST00000288602', '1..1000')
        """
        self._check_frmt(frmt, ["json"])
        res = self.services.http_get(
            "map/cds/{0}/{1}".format(identifier, region),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback, "species": species},
        )

        return res

    def get_map_cdna_to_region(self, identifier, region, frmt="json", species=None):
        """Convert from cDNA coordinates to genomic coordinates.

        :param str first: version of the input assembly
        :param str first: version of the output assembly
        :param str region:  query region (see example)
        :param str species: default to human

        Output reflects forward orientation coordinates as returned from the Ensembl API.


        ::

            get_map_cdna_to_region('ENST00000288602', '100..300')

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "map/cdna/{0}/{1}".format(identifier, region),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback, "species": species},
        )
        return res

    def get_map_assembly_one_to_two(self, first, second, region, species="human", frmt="json"):
        """Convert the co-ordinates of one assembly to another


        :param str first: version of the input assembly
        :param str first: version of the output assembly
        :param str region:  query region (see example)
        :param str species: default to human

        ::

            e.get_map_assembly_one_to_two(species='human',
                first='GRCh37', region='X:1000000..1000100:1', second='GRCh38')

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "map/{0}/{1}/{2}/{3}".format(species, first, region, second),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={},
        )
        return res

    def get_map_translation_to_region(self, identifier, region, frmt="json", species=None):
        """Convert from protein (translation) coordinates to genomic coordinates.

        Output reflects forward orientation coordinates as returned from the
        Ensembl API.

        :param identifier: a stable Ensembl ID
        :param str query: a query region
        :param str species: Species name/alias (e.g., homo_sapiens)

        ::

            get_map_translation_to_region('ENSP00000288602', '100..300')
        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "map/translation/{0}/{1}".format(identifier, region),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    # ONTOLOGY and Taxonomy
    # -------------------------------------------------------------------------

    def get_ontology_by_id(self, identifier, frmt="json", relation=None, simple=False):
        """Search for an ontological  term by its namespaced identifier

        :param str identifier: An ontology term identifier (e.g., GO:0005667)
        :param bool simple: If set the API will avoid the fetching of parent and child terms
        :param str frmt: response formats in json, xml, yaml, jsonp
        :param str simple: If set the API will avoid the fetching of parent and child terms
        :param str relation: The types of relationships to include in the output. Fetches
            all relations by default (e.g., is_a, part_of)

        ::

            >>> from bioservices imoprt Ensembl
            >>> e = Ensembl()
            >>> res = e.get_ontology('GO:0005667')

        """
        self._check_frmt(frmt, ["xml", "yaml"])
        identifier = str(identifier)
        res = self.services.http_get(
            "ontology/id/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "simple": int(simple),
                "relation": relation,
                "callback": self.callback,
            },
        )
        return res

    def get_ontology_by_name(self, name, frmt="json", ontology=None, relation=None, simple=False):
        """Search for a list of ontological terms by their name

        :param str name: An ontology name. SQL wildcards See :class:`Ensembl` doc.
        :param str frmt: response formats in json, xml, yaml, jsonp
        :param str simple: If set the API will avoid the fetching of parent and child terms
        :param str relation: The types of relationships to include in the output. Fetches
            all relations by default (e.g., is_a, part_of)
        :param str ontology: Filter by ontology. Used to disambiguate terms which are
            shared between ontologies such as GO and EFO (e.g., GO)


        ::

            >>> from bioservices imoprt Ensembl
            >>> e = Ensembl()
            >>> res = e.get_ontology_by_name('transcription factor')
            400
            >>> res = e.get_ontology_by_name('transcription factor complex')
            >>> res[0]['children']

        """
        self._check_frmt(frmt, ["xml", "yaml"])
        res = self.services.http_get(
            "ontology/name/{0}".format(name),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "simple": int(simple),
                "relation": relation,
                "callback": self.callback,
                "ontology": ontology,
            },
        )
        return res

    def get_taxonomy_by_id(self, identifier, frmt="json", simple=False):
        """Search for a taxonomic term by its identifier or name

        :param str identifier: A taxon identifier. Can be a NCBI taxon id or
            a name (e.g., 9606 or Homo sapiens)
        :param bool simple: If set the API will avoid the fetching of parent and child terms
        :param str frmt: response formats in json, xml, yaml, jsonp
        """
        self._check_frmt(frmt, ["xml", "yaml"])
        res = self.services.http_get(
            "taxonomy/id/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"simple": int(simple)},
        )
        return res

    def get_taxonomy_by_name(self, name, frmt="json"):
        """Search for a taxonomic id by a non-scientific name

        :param str name: A non-scientific species name. Can include SQL wildcards
            See :class:`Ensembl` doc.
        :param str frmt: response formats in json, xml, yaml, jsonp

        ::

            >>> from bioservices imoprt Ensembl
            >>> e = Ensembl()
            >>> res = e.get_taxonomy_by_name('homo')

        """
        self._check_frmt(frmt, ["xml", "yaml"])
        res = self.services.http_get(
            "taxonomy/name/{0}".format(name),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    def get_taxonomy_classification_by_id(
        self,
        identifier,
        frmt="json",
    ):
        """Return the taxonomic classification of a taxon node


        :param str identifier: A taxon identifier. Can be a NCBI taxon id or a name
            (e.g., 9606, Homo sapiens)
        :param str frmt: json, xml, yaml, jsonp

        ::

            >>> from bioservices imoprt Ensembl
            >>> e = Ensembl()
            >>> res = e.get_taxonomy_classification_by_id('9606')

        """
        self._check_frmt(frmt, ["xml", "yaml"])
        res = self.services.http_get(
            "taxonomy/classification/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    def get_ontology_ancestors_by_id(self, identifier, frmt="json", ontology=None):
        """Reconstruct the entire ancestry of a term from is_a and part_of
        relationships

        :param str identifier: An ontology term identifier (e.g., GO:0005667)
        :param str frmt: json, xml, yaml, jsonp
        :param str ontology: Filter by ontology. Used to disambiguate
            terms which are shared between ontologies such as GO and EFO (e.g.,
            GO)


        """
        self._check_frmt(frmt, ["xml", "yaml"])
        res = self.services.http_get(
            "ontology/ancestors/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback, "ontology": ontology},
        )
        return res

    def get_ontology_ancestors_chart_by_id(self, identifier, frmt="json", ontology=None):
        """Reconstruct the entire ancestry of a term from is_a and part_of
        relationships.

        :param str identifier: an ontology term identifier (GO:0005667)
        :param str frmt: json, xml, yaml, jsonp
        :param str ontology: Filter by ontology. Used to disambiguate
            terms which are shared between ontologies such as GO and EFO

        """
        self._check_frmt(frmt, ["xml", "yaml"])
        res = self.services.http_get(
            "ontology/ancestors/chart/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback, "ontology": ontology},
        )
        return res
        #'getAncestorsChartById': {'url': '/ontology/ancestors/chart/{{id}}',

    def get_ontology_descendants_by_id(
        self,
        identifier,
        frmt="json",
        closest_term=None,
        ontology=None,
        subset=None,
        zero_distance=None,
    ):
        """Find all the terms descended from a given term. By default searches
        are conducted within the namespace of the given identifier

        :param str identifier: an ontology term identifier (GO:0005667)
        :param str frmt: json, xml, jsonp
        :param bool closest_term: If true return only the closest terms to the
            specified term
        :param str ontology: Filter by ontology. Used to disambiguate terms
            which are shared between ontologies such as GO and EFO
        :param str subset: Filter terms by the specified subset
        :param bool zero_distance: Return terms with a distance of 0

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "ontology/descendants/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "callback": self.callback,
                "ontology": ontology,
                "closest_term": closest_term,
                "subset": subset,
                "zero_distance": zero_distance,
            },
        )
        return res

    # OVERLAP
    # -----------------------------------------------------
    def get_overlap_by_id(
        self,
        identifier,
        feature=None,
        frmt="json",
        biotype=None,
        db_type=None,
        logic_name=None,
        misc_set=None,
        object_type=None,
        so_term=None,
        species=None,
        species_set="mammals",
    ):
        """Retrieves features (e.g. genes, transcripts, variations etc.)
        that overlap a region defined by the given identifier.

        :param str identifier: An Ensemble stable ID
        :param str feature: The type of feature to retrieve. Multiple values
            are accepted. Value in Enum(gene, transcript, cds, exon, repeat,
            simple, misc, variation, somatic_variation, structural_variation,
            somatic_structural_variation, constrained, regulatory
        :param str biotype: The functional classification of the gene or
            transcript to fetch. Cannot be used in conjunction with logic_name
            when querying transcripts. (e.g., protein_coding)
        :param str db_type: Restrict the search to a database other than
            the default. Useful if you need to use a DB other than core
        :param str logic_name: Limit retrieval of genes, transcripts and
            exons by a given name of an analysis.
        :param str misc_set: Miscellaneous set which groups together
            feature entries. Consult the DB or returned data sets to discover
            what is available. (e.g., cloneset_30k
        :param str object_type: Filter by feature type (e.g., gene)
        :param str so_term: Sequence Ontology term to narrow down the
            possible variations returned. (e.g., SO:0001650)
        :param str species: Species name/alias.
        :param str species_set: Filter by species set for
            retrieving constrained elements. (e.g. mammals)

        """
        self._check_frmt(frmt, ["xml", "gff3", "bed"])
        res = self.services.http_get(
            "overlap/id/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "callback": self.callback,
                "biotype": biotype,
                "db_type": db_type,
                "logic_name": logic_name,
                "misc_set": misc_set,
                "object_type": object_type,
                "so_term": so_term,
                "species": species,
                "feature": feature,
                "species_set": species_set,
            },
        )
        return res

    def get_overlap_by_region(
        self,
        region,
        species,
        feature=None,
        frmt="json",
        biotype=None,
        cell_type=None,
        db_type=None,
        logic_name=None,
        misc_set=None,
        object_type=None,
        so_term=None,
        species_set=None,
        trim_downstream=False,
        trim_upstream=False,
    ):
        """Retrieves multiple types of features for a given region.

        :param str region: Query region. A maximum of 5Mb is allowed to
            be requested at any one time. e.g.,  X:1..1000:1, X:1..1000:-1,
            X:1..1000
        :param str species: Species name/alias.
        :param str feature: The type of feature to retrieve. Multiple
            values are accepted: gene, transcript, cds, exon, repeat,
            simple, misc, variation, somatic_variation, structural_variation,
            somatic_structural_variation, constrained, regulatory
        :param str biotype: The functional classification of the gene or
            transcript to fetch. Cannot be used in conjunction with logic_name
            when querying transcripts. (e.g., protein_coding)
        :param cell_type: Cell type name in Ensembl's Regulatory Build,
            required for segmentation feature, optional for regulatory elements.
            e.g., K562
        :param str db_type: Restrict the search to a database other than
            the default. Useful if you need to use a DB other than core
        :param str logic_name: Limit retrieval of genes, transcripts and
            exons by a given name of an analysis.
        :param str misc_set: Miscellaneous set which groups together
            feature entries. Consult the DB or returned data sets to discover
            what is available. (e.g., cloneset_30k)
        :param str so_term: Sequence Ontology term to narrow down the
            possible variations returned. (e.g., SO:0001650)
        :param str species_set: Filter by species set for
            retrieving constrained elements. (e.g. mammals)
        :param bool trim_downstream: Do not return features which overlap
            the downstream end of the region.
        :param bool trim_upstream: Do not return features which overlap
            upstream end of the region.

        .. todo:: feature can take several values. how can be do that.
        """
        self._check_frmt(frmt, ["xml", "gff3", "bed"])
        res = self.services.http_get(
            "overlap/region/{0}/{1}".format(species, region),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "callback": self.callback,
                "biotype": biotype,
                "cell_type": cell_type,
                "feature": feature,
                "db_type": db_type,
                "logic_name": logic_name,
                "misc_set": misc_set,
                "so_term": so_term,
                "species_set": species_set,
                "trim_downstream": trim_downstream,
                "trim_upstream": trim_upstream,
            },
        )
        return res

    def get_overlap_by_translation(
        self,
        identifier,
        frmt="json",
        db_type=None,
        feature="protein_feature",
        so_term=None,
        species=None,
        type="none",
    ):
        """Retrieve features related to a specific Translation as
        described by its stable ID (e.g. domains, variations).

        :param str identifier:
        :param str frmt:
        :param str db_type: Restrict the search to a database other
            than the default. Useful if you need to use a DB other than core
        :param str feature: requested feature in: transcript_variation,
            protein_feature, residue_overlap, translation_exon,
            somatic_transcript_variation
        :param str so_term: Sequence Ontology term to restrict the variations
            found. Its descendants are also included in the search. (e.g.,
            SO:0001650)
        :param str species: species name/alias
        :param str type: Type of data to filter by. By default, all
            features are returned. Can specify a domain or consequence
            type. (e.g.,  low_complexity)
        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "overlap/translation/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "callback": self.callback,
                "db_type": db_type,
                "feature": feature,
                "so_term": so_term,
                "species": species,
                "type": type,
            },
        )
        return res

    # REGULATION
    # --------------------------------------
    def get_regulatory_by_id(
        self,
        identifier,
        species,
        frmt="json",
    ):
        """Returns a RegulatoryFeature given its stable ID

        :param str identifier:
        :param str species:


        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "regulatory/{0}/{1}".format(species, identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={"callback": self.callback},
        )
        return res

    # SEQUENCE
    # -----------------------------
    def get_sequence_by_id(
        self,
        identifier,
        frmt="fasta",
        db_type=None,
        expand_3prime=None,
        expand_5prime=None,
        format=None,
        mask=None,
        mask_feature=False,
        multiple_sequences=False,
        object_type=None,
        species=None,
        type="genomic",
    ):
        """Request multiple types of sequence by stable identifier.

        :param str identifier:
        :param str frmt: response formats: fasta, json, text, yam, jsonp
        :param str db_type: Restrict the search to a database other than the
            default. Useful if you need to use a DB other than core (e.g.,
            core)
        :param int expand_3prime: Expand the sequence downstream of the
            sequence by this many basepairs. Only available when using
            genomic sequence type.
        :param int expand_5prime: Expand the sequence upstream of the
            sequence by this many basepairs. Only available when using
            genomic sequence type.
        :param str format: Format of the data (e.g., fasta)
        :param str mask: Request the sequence masked for repeat sequences.
            Hard will mask all repeats as N's and soft will mask repeats
            as lowercased characters. Only available when using genomic
            sequence type. (hard/soft)
        :param bool mask_feature: Mask features on the sequence. If sequence
            is genomic, mask introns. If sequence is cDNA, mask UTRs.
            Incompatible with the 'mask' option
        :param bool multiple_sequences: Allow the service to return more
            than 1 sequence per identifier. This is useful when querying
            for a gene but using a type such as protein.
        :param str object_type: Filter by feature type (e.g., gene)
        :param str species: Species name/alias (e.g., homo_sapiens)
        :param str type:. could be genomic, cds, cdna, protein (homo_sapiens).
            Requesting a gene and kind not equal to genomic may result in
            multiple sequence, which required the parameter multi_sequences
            to be set to True

        ::

            >>> # Default format is fasta, let us use parameter frmt to overwrite it
            >>> sequence = e.get_sequence('ENSG00000157764', frmt='text')
            >>> print(sequence[0:10])
            CGCCTCCCTTCCCCCTCCCC

            >>> # complex request for different database and kind
            >>> res = e.get_sequence('CCDS5863.1', frmt='fasta',
                    object_type='transcript', db_type='otherfeatures',
                    type='cds', species='human')
            >>> print(res[0:100])
            >CCDS5863.1
            ATGGCGGCGCTGAGCGGTGGCGGTGGTGGCGGCGCGGAGCCGGGCCAGGCTCTGTTCAAC
            GGGGACATGGAGCCCGAGGCCGGCGCC

            >>>


        """
        self._check_frmt(frmt, ["fasta", "text", "yaml", "seqxml"])
        multiple_sequences = int(multiple_sequences)
        res = self.services.http_get(
            "sequence/id/{0}".format(identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "db_type": db_type,
                "object_type": object_type,
                "multiple_sequences": multiple_sequences,
                "species": species,
                "expand_3prime": expand_3prime,
                "expand_5prime": expand_5prime,
                "format": format,
                "mask": mask,
                "mask_feature": mask_feature,
                "type": type,
                "multiple_sequences": multiple_sequences,
                "species": species,
            },
        )
        return res

    def get_sequence_by_region(
        self,
        region,
        species,
        frmt="json",
        coord_system=None,
        coord_system_version=None,
        expand_3prime=None,
        expand_5prime=None,
        format=None,
        mask=None,
        mask_feature=False,
    ):
        """Returns the genomic sequence of the specified region of the given species.

        :param str region: Query region. A maximum of 10Mb is allowed to be
            requested at any one time. e.g.,  X:1000000..1000100:1
        :param str species: Species name/alias
        :param str coord_system: Filter by coordinate system name (e.g., contig,
            seqlevel)
        :param str  coord_system_version: Filter by coordinate system version
            (e.g., GRCh37)
        :param int expand_3prime: Expand the sequence downstream of the
            sequence by this many basepairs. Only available when using
            genomic sequence type.
        :param int expand_5prime: Expand the sequence upstream of the
            sequence by this many basepairs. Only available when using
            genomic sequence type.
        :param str format: Format of the data. (e.g., fasta)
        :param str mask: Request the sequence masked for repeat sequences.
            Hard will mask all repeats as N's and soft will mask repeats
            as lowercased characters. Only available when using genomic
            sequence type. (hard/soft)
        :param bool mask_feature: Mask features on the sequence. If sequence
            is genomic, mask introns. If sequence is cDNA, mask UTRs.
            Incompatible with the 'mask' option

        """
        self._check_frmt(frmt, ["fasta", "text", "yaml", "seqxml"])
        res = self.services.http_get(
            "sequence/region/{0}/{1}".format(species, region),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "callback": self.callback,
                "coord_system": coord_system,
                "coord_system_version": coord_system_version,
                "expand_3prime": expand_3prime,
                "expand_5prime": expand_5prime,
                "format": format,
                "mask": mask,
                "mask_feature": mask_feature,
            },
        )
        return res

    # VARIATION
    # -----------------------------------------------------
    def get_variation_by_id(
        self,
        identifier,
        species,
        frmt="json",
        genotypes=False,
        phenotypes=False,
        pops=False,
    ):
        """

        :param str identifier: variation identifier (e.g., rs56116432)
        :param str species: Species name/alias (e.g., homo_sapiens)
        :param str frmt: response format (json, xml, jsonp)
        :param bool genotypes: Include genotypes
        :param bool phenotypes: Include phenotypes
        :param bool pops: Include populations

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "variation/{0}/{1}".format(species, identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "genotypes": int(genotypes),
                "phenotypes": int(phenotypes),
                "pops": int(pops),
            },
        )
        return res

    def get_vep_by_id(
        self,
        identifier,
        species,
        frmt="json",
        canonical=False,
        ccds=False,
        domains=False,
        hgvs=False,
        numbers=False,
        protein=False,
        xref_refseq=False,
    ):
        """Fetch variant consequences based on a variation identifier

        :param str identifier: Query ID. Supports dbSNP, COSMIC and
            HGMD identifiers   (e.g.,    rs116035550, COSM476)
        :param str species: Species name/alias
        :param bool canonical: Include a flag indicating the canonical transcript for a gene
        :param bool ccds: Include CCDS transcript identifiers
        :param bool domains:Include names of overlapping protein domains
        :param bool hgvs: Include HGVS nomenclature based on Ensembl stable identifiers
        :param bool numbers: Include affected exon and intron positions within the transcript
        :param bool protein: Include Ensembl protein identifiers
        :param bool xref_refseq: Include aligned RefSeq mRNA identifiers for
            transcript. NB: theRefSeq and Ensembl transcripts aligned in this
            way MAY NOT, AND FREQUENTLY WILL NOT, match exactly in sequence,
            exon structure and protein product
        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "vep/{0}/id/{1}".format(species, identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "callback": self.callback,
                "canonical": canonical,
                "ccds": ccds,
                "domains": domains,
                "hgvs": hgvs,
                "numbers": numbers,
                "protein": protein,
                "xref_refseq": xref_refseq,
            },
        )
        return res

    def post_vep_by_id(self, species, identifiers):
        raise NotImplementedError
        # POST vep/:species/id/   Fetch variant consequences for multiple ids
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "variation/{0}/{1}".format(species, identifier),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "genotypes": int(genotypes),
                "phenotypes": int(phenotypes),
                "pops": int(pops),
            },
        )
        raise NotImplementeError

    def get_vep_by_region(
        self,
        region,
        allele,
        species,
        frmt="json",
        canonical=False,
        ccds=False,
        domains=False,
        hgvs=False,
        numbers=False,
        protein=False,
        xref_refseq=False,
    ):
        """Fetch variant consequences

        :param region: Query region e.g,  9:22125503-22125502:1
        :param str allele: Variation allele (e.g., C, DUP)
        :param str species: Species name/alias
        :param bool canonical: Include a flag indicating the canonical transcript for a gene
        :param bool ccds: Include CCDS transcript identifiers
        :param bool domains:Include names of overlapping protein domains
        :param bool hgvs: Include HGVS nomenclature based on Ensembl stable identifiers
        :param bool numbers: Include affected exon and intron positions within the transcript
        :param bool protein: Include Ensembl protein identifiers
        :param bool xref_refseq: Include aligned RefSeq mRNA identifiers for
            transcript. NB: theRefSeq and Ensembl transcripts aligned in this
            way MAY NOT, AND FREQUENTLY WILL NOT, match exactly in sequence,
            exon structure and protein product

        """
        self._check_frmt(frmt, ["xml"])
        res = self.services.http_get(
            "vep/{0}/region/{1}/{2}".format(species, region, allele),
            frmt=frmt,
            headers=self.services.get_headers(content=frmt),
            params={
                "callback": self.callback,
                "canonical": canonical,
                "ccds": ccds,
                "domains": domains,
                "hgvs": hgvs,
                "numbers": numbers,
                "protein": protein,
                "xref_refseq": xref_refseq,
            },
        )
        return res

    def post_vep_by_region(self, species, region):
        # POST vep/:species/region/   Fetch variant consequences for multiple regions
        raise NotImplementeError


class EnsemblFTP(object):
    """Draft to be completed."""

    def __init__(self):
        pass

    def download(self, db="fungi", release="current"):

        import ftplib

        f = ftplib.FTP("ftp.ensemblgenomes.org")
        f.login("anonymous", "anonymous")
        f.cwd("pub/%s/%s" % (db, release))
        f.cwd("fasta")
        species = f.nlst()

        print(species)

        for this in species:
            print(this)
            if this.endswith("collection"):
                f.cwd(this)
                subdirs = f.nlst()
                for thisdir in subdirs:
                    print(thisdir)
                    f.cwd(thisdir)
                    f.cwd("dna")
                    files = f.nlst()
                    todownload = [x for x in files if x.endswith("dna.genome.fa.gz")]
                    for filename in todownload:
                        f.retrbinary("RETR %s" % filename, open(filename, "wb").write)
                    f.cwd("../../")
                f.cwd("..")

            else:
                continue
                f.cwd(this)
                f.cwd("dna")
                files = f.nlst()
                todownload = [x for x in files if x.endswith("dna.genome.fa.gz")]
                for filename in todownload:
                    f.retrbinary("RETR %s" % filename, open(filename, "wb").write)
                f.cwd("../../")



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      https://github.com/cokelaer/bioservices
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  source: http://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################

"""Interface to PRIDE web service

.. topic:: What is PRIDE ?

    :URL: http://www.ebi.ac.uk/pride/ws/archive/v2

    .. highlights::

         The PRIDE PRoteomics IDEntifications database is a centralized,
         standards compliant, public data repository for proteomics data,
         including protein and peptide identifications, post-translational
         modifications and supporting spectral evidence.

        -- From PRIDE web site, Jan 2015


"""
import tqdm

from bioservices.services import REST
from bioservices import logger

logger.name = __name__


__all__ = ["PRIDE"]


class PRIDE:
    """Interface to the `PRIDE <http://rest.ensembl.org>`_ service



    ::

        from bioservices import PRIDE
        p = PRIDE()
        p.get_peptide_evidence(projectAccession)

    .. versionchanged:: 1.10.1

        Due to new API:

        - the method project_count was dropped.
        - get_project_list was renamed in get_project_files
        - get_assays, get_assay_count, get_assay_count_project_accession, get_assay_list were dropped in v2
        - get_protein_list, get_protein_count, get_protein_count_assay, get_protein_list, get_protein_list_assay
          replaced by get_protein_evidences method
        - get_peptide_list_assay, get_peptide_count, get_peptide_list, get_peptide_list_sequence,
          get_peptide_count_assay replaced by get_peptide_evidence.

    """

    _url = "https://www.ebi.ac.uk/pride/ws/archive/v2"

    def __init__(self, verbose=False, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        :param cache: set to True to use caching. Not recommended for
            this service that evolves a lot
        """
        self.services = REST(name="PRIDE", url=PRIDE._url, verbose=verbose, cache=cache)

    def get_project(self, identifier):
        """Retrieve project information by accession

        List of PRIDE Archive Projects. The following method do not allows
        to perform search, for search functionality you will need to use
        the search/projects. The result list is Paginated using the pageSize and page.

        :param str identifier: a valid PRIDE identifier e.g., PRD000001

        :return: if identifier is invalid, returns an emppty dictionary {}

        .. doctest::

            >>> from bioservices import PRIDE
            >>> p = PRIDE()
            >>> res = p.get_project("PRD000001")
            >>> res['title']
            'COFRADIC proteome of unstimulated human blood platelets'

        """
        res = self.services.http_get(f"projects/{identifier}")
        if res == 400:
            logger.warning(
                f"Nothing found for {identifier}. may be this is not a valid identifier. Use get_projects"
            )
            return {}
        return res

    def get_projects(self, pageSize=100, max_pages=1e9):
        """Get list of all projects"""
        N = self.get_projects_count()
        Npages = int(N / pageSize)
        Npages = min(Npages, max_pages)

        results = []
        page_count = 0
        for page in tqdm.tqdm(range(Npages)):
            res = self.services.http_get("projects", params={"pageSize": pageSize, "page": page})
            results.extend(res["_embedded"]["projects"])
            page_count += 1
            if page_count > max_pages:
                break

        return results

    def get_projects_count(self):
        res = self.services.http_get("projects")
        return res["page"]["totalElements"]

    def get_project_files(self, accession, pageSize=100, page=0, sortConditions=None, sortDirection="DESC", filters=""):
        """list projects or given criteria

        :param str accession: the accession number to look for
        :param int pageSize: how many results to return per page
        :param int page: which page (starting from 0) of the result to return
        :param str sortConditions: default is submission_date but more fields
            can be separated by comma and passed. Example: submission_date,project_title
        :param str sortDirection: the sorting order (ASC or DESC)
        :param str filters: Parameters to filter the search results. The structure of
            the filter is: field1==value1, field2==value2. Example accession==PRD000001

        ::

            >>> p = PRIDE()
            >>> results = p.get_project_files(accession="PRD000001", pageSize=10, page=1)


        In v1.10.1 due to new PRIDE API, the method **get_file_count** was dropped. You can use::

            len(results['_embedded']['files'])

        Similarly the **get_file_list** method was dropped since all results are
        stored in the output of this method


        """
        params = {
            "pageSize": pageSize,
            "page": page,
            "sortDirection": sortDirection,
            "sortConditions": sortConditions,
            "filter": filters,
        }

        res = self.services.http_get(f"projects/{accession}/files", params=params)
        try:
            res = res["list"]
        except:
            pass
        return res

    def get_protein_evidences(
        self,
        project_accession=None,
        assay_accession=None,
        reported_accession=None,
        pageSize=100,
        page=0,
        sortDirection="DESC",
        sortConditions="projectAccession",
    ):

        """Get all proteins evidence

        :param project_accession:
        :param assay_accession:
        :param reported_accession:
        :param int pageSize: how many results to return per page
        :param int page: which page (starting from 0) of the result to return
        :param str sortConditions: default is submission_date but more fields
            can be separated by comma and passed. Example: submission_date,project_title
        :param str sortDirection: the sorting order (ASC or DESC)

        ::

            p.get_protein_evidences()['_embedded']['proteinevidences']
        """

        params = {}
        if project_accession:
            params["projectAccession"] = project_accession
        if assay_accession:  # pragma: no cover
            params["assayAccession"] = assay_accession
        if reported_accession:  # pragma: no cover
            params["reportedAccession"] = reported_accession
        params["pageSize"] = pageSize
        params["page"] = page
        params["sortConditions"] = sortConditions
        params["sortDirection"] = sortDirection

        res = self.services.http_get(f"proteinevidences", params=params)
        return res

    def get_peptide_evidence(
        self,
        project_accession=None,
        assay_accession=None,
        protein_accession=None,
        peptide_evidence_accession=None,
        peptide_sequence=None,
        pageSize=100,
        page=0,
        sortDirection="DESC",
        sortConditions="projectAccession",
    ):
        """Get all the peptide evidences for an specific protein evidence


        :param project_accession:
        :param assay_accession:
        :param protein_accession:
        :param peptide_evidence_accession:
        :param peptide_sequence:
        :param int pageSize: how many results to return per page
        :param int page: which page (starting from 0) of the result to return
        :param str sortConditions: default is submission_date but more fields
            can be separated by comma and passed. Example: submission_date,project_title
        :param str sortDirection: the sorting order (ASC or DESC)

        Retrieving data from project accession should be fast::

            p.get_peptide_evidence(protein_accession="Q8IX30")

        but other methods may be slow::

            p.get_peptide_evidence(peptide_sequence="CQGSPGASKAMLSCNR")
        """
        params = {}
        if project_accession:
            params["projectAccession"] = project_accession
        if assay_accession:  # pragma: no cover
            params["assayAccession"] = assay_accession
        if protein_accession:  # pragma: no cover
            params["proteinAccession"] = protein_accession
        if peptide_evidence_accession:  # pragma: no cover
            params["peptideEvidenceAccession"] = peptide_evidence_accession
        if peptide_sequence:  # pragma: no cover
            params["peptideSequence"] = peptide_sequence
        params["pageSize"] = pageSize
        params["page"] = page
        params["sortConditions"] = sortConditions

        res = self.services.http_get(f"peptideevidences", params=params)
        return res

    def get_stats(self, name=None):
        """Retrieve statistics by Name

        If you do not have the name, just type::

            p.get_stats()

        and then, e.g., ::

            p.get_stats("SUBMISSIONS_PER_YEAR")

        """

        if name is None:
            res = self.services.http_get("stats/")
        else:
            res = self.services.http_get(f"stats/{name}")
        return res



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      https://github.com/cokelaer/bioservices
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  source: http://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Modules with common tools to access web resources"""
import os
import sys
import time
import socket
import platform
import traceback

from bioservices.settings import BioServicesConfig

# fixing compatiblity python 2 and 3 related to merging or urllib and urllib2 in python 3
try:
    # python 3
    from urllib.request import urlopen
    from urllib.parse import urlparse, urlencode
    from urllib.error import HTTPError
    from urllib.request import Request
except:
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError

# fixing compatibility issue of input/raw_input
if "raw_input" in __builtins__:
    input = raw_input

# Indded, we want suds_jurko instead
sys.path = [x for x in sys.path if "suds-" not in x]


from easydev import DevTools
import colorlog


__all__ = ["Service", "WSDLService", "BioServicesError", "REST"]


class BioServicesError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Service:
    """Base class for WSDL and REST classes

    .. seealso:: :class:`REST`, :class:`WSDLService`
    """

    #: some useful response codes
    response_codes = {
        200: "OK",
        201: "Created",
        400: "Bad Request. There is a problem with your input",
        404: "Not found. The resource you requests does not exist",
        405: "Method not allowed",
        406: "Not Acceptable. Usually headers issue",
        410: "Gone. The resource you requested was removed.",
        415: "Unsupported Media Type",
        500: "Internal server error. Most likely a temporary problem",
        503: "Service not available. The server is being updated, try again later",
    }

    def __init__(self, name, url=None, verbose=True, requests_per_sec=10, url_defined_later=False):
        """.. rubric:: Constructor

        :param str name: a name for this service
        :param str url: its URL
        :param bool verbose: prints informative messages if True (default is
            True)
        :param requests_per_sec: maximum number of requests per seconds
            are restricted to 3. You can change that value. If you reach the
            limit, an error is raise. The reason for this limitation is
            that some services (e.g.., NCBI) may black list you IP.
            If you need or can do more (e.g., ChEMBL does not seem to have
            restrictions), change the value. You can also have several instance
            but again, if you send too many requests at the same, your future
            requests may be retricted. Currently implemented for REST only


        All instances have an attribute called :attr:`~Service.logging` that
        is an instanceof the :mod:`logging` module. It can be used to print
        information, warning, error messages::

            self.logging.info("informative message")
            self.logging.warning("warning message")
            self.logging.error("error message")

        The attribute :attr:`~Service.debugLevel`  can be used to set the behaviour
        of the logging messages. If the argument verbose is True, the debugLebel
        is set to INFO. If verbose if False, the debugLevel is set to WARNING.
        However, you can use the :attr:`debugLevel` attribute to change it to
        one of DEBUG, INFO, WARNING, ERROR, CRITICAL. debugLevel=WARNING means
        that only WARNING, ERROR and CRITICAL messages are shown.

        """
        super(Service, self).__init__()
        self.requests_per_sec = requests_per_sec
        self.name = name
        self.logging = colorlog.getLogger(f"bioservices.{self.name}")

        if verbose:
            self.logging.setLevel("INFO")
        else:
            self.logging.setLevel("WARNING")

        self._url = url
        try:
            if self.url is not None:
                urlopen(self.url)
        except Exception as err:
            if url_defined_later is False:
                self.logging.warning("The URL (%s) provided cannot be reached." % self.url)
        self._easyXMLConversion = True

        self.devtools = DevTools()
        self.settings = BioServicesConfig()

        self._last_call = 0

    def _calls(self):
        time_lapse = 1.0 / self.requests_per_sec
        current_time = time.time()
        dt = current_time - self._last_call

        if self._last_call == 0:
            self._last_call = current_time
            return
        else:
            self._last_call = current_time
            if dt > time_lapse:
                return
            else:
                time.sleep(time_lapse - dt)

    def _get_caching(self):
        return self.settings.params["cache.on"][0]

    def _set_caching(self, caching):
        self.devtools.check_param_in_list(caching, [True, False])
        self.settings.params["cache.on"][0] = caching
        # reset the session, which will be automatically created if we
        # access to the session attribute
        self._session = None

    CACHING = property(_get_caching, _set_caching)

    def _get_url(self):
        return self._url

    def _set_url(self, url):
        # something more clever here to check the URL e.g. starts with http
        if url is not None:
            url = url.rstrip("/")
            self._url = url

    url = property(_get_url, _set_url, doc="URL of this service")

    def _get_easyXMLConversion(self):
        return self._easyXMLConversion

    def _set_easyXMLConversion(self, value):
        if isinstance(value, bool) is False:
            raise TypeError("value must be a boolean value (True/False)")
        self._easyXMLConversion = value

    easyXMLConversion = property(
        _get_easyXMLConversion,
        _set_easyXMLConversion,
        doc="""If True, xml output from a request are converted to easyXML object (Default behaviour).""",
    )

    def easyXML(self, res):
        """Use this method to convert a XML document into an
            :class:`~bioservices.xmltools.easyXML` object

        The easyXML object provides utilities to ease access to the XML
        tag/attributes.

        Here is a simple example starting from the following XML

        .. doctest::

            >>> from bioservices import *
            >>> doc = "<xml> <id>1</id> <id>2</id> </xml>"
            >>> s = Service("name")
            >>> res = s.easyXML(doc)
            >>> res.findAll("id")
            [<id>1</id>, <id>2</id>]

        """
        from bioservices import xmltools

        return xmltools.easyXML(res)

    def __str__(self):
        txt = "This is an instance of %s service" % self.name
        return txt

    def pubmed(self, Id):
        """Open a pubmed Id into a browser tab

        :param Id: a valid pubmed Id in string or integer format.

        The URL is a concatenation of the pubmed URL
        http://www.ncbi.nlm.nih.gov/pubmed/ and the provided Id.

        """
        url = "http://www.ncbi.nlm.nih.gov/pubmed/"
        import webbrowser

        webbrowser.open(url + str(Id))

    def on_web(self, url):
        """Open a URL into a browser"""
        import webbrowser

        webbrowser.open(url)

    def save_str_to_image(self, data, filename):
        """Save string object into a file converting into binary"""
        with open(filename, "wb") as f:
            import binascii

            try:
                # python3
                newres = binascii.a2b_base64(bytes(data, "utf-8"))
            except:
                newres = binascii.a2b_base64(data)
            f.write(newres)


class WSDLService(Service):
    """Class dedicated to the web services based on WSDL/SOAP protocol.

    .. seealso:: :class:`RESTService`, :class:`Service`

    """

    _service = "WSDL"

    def __init__(self, name, url, verbose=True, cache=False):
        """.. rubric:: Constructor

        :param str name: a name e.g. Kegg, Reactome, ...
        :param str url: the URL of the WSDL service
        :param bool verbose: prints informative messages

        The :attr:`serv` give  access to all WSDL functionalities of the service.

        The :attr:`methods` is an alias to self.serv.methods and returns
        the list of functionalities.

        """
        super(WSDLService, self).__init__(name, url, verbose=verbose)

        self.logging.info("Initialising %s service (WSDL)" % self.name)
        self.CACHING = cache

        try:
            #: attribute to access to the methods provided by this WSDL service
            from suds.client import Client
            from suds.cache import ObjectCache

            oc = ObjectCache(self.settings.user_config_dir, days=0)
            if self.CACHING is True:
                self.suds = Client(self.url, cache=oc, cachingpolicy=1)
            else:
                self.suds = Client(self.url)
            # reference to the service
            self.serv = self.suds.service
            self._update_settings()
        except Exception:
            self.logging.error("Could not connect to the service %s " % self.url)
            raise Exception

    def _update_settings(self):
        self.TIMEOUT = self.settings.TIMEOUT

    def wsdl_methods_info(self):
        methods = self.suds.wsdl.services[0].ports[0].methods.values()
        for method in methods:
            try:
                print(
                    "%s(%s) "
                    % (
                        method.name,
                        ", ".join(
                            "type:%s: %s - element %s" % (part.type, part.name, part.element)
                            for part in method.soap.input.body.parts
                        ),
                    )
                )
            except:
                print(method)

    def _get_methods(self):
        return [x.name for x in self.suds.wsdl.services[0].ports[0].methods.values()]

    wsdl_methods = property(_get_methods, doc="returns methods available in the WSDL service")

    def wsdl_create_factory(self, name, **kargs):
        params = self.suds.factory.create(name)

        # e.g., for eutils
        if "email" in dict(params).keys():
            params.email = self.settings.params["user.email"][0]

        if "tool" in dict(params).keys():
            import bioservices

            params.tool = "BioServices, " + bioservices.__version__

        for k, v in kargs.items():
            from suds import sudsobject

            keys = sudsobject.asdict(params).keys()
            if k in keys:
                params[k] = v
            else:
                msg = "{0} incorrect. Correct ones are {1}"
                self.logging.error(msg.format(k, keys))
        return params

    def _get_timeout(self):
        return self.suds.options.timeout

    def _set_timeout(self, value):
        self.suds.set_options(timeout=value)
        self.settings.TIMEOUT = value

    TIMEOUT = property(_get_timeout, _set_timeout)


class RESTbase(Service):
    _service = "REST"

    def __init__(self, name, url=None, verbose=True, requests_per_sec=3, url_defined_later=False):
        super(RESTbase, self).__init__(
            name,
            url,
            verbose=verbose,
            requests_per_sec=requests_per_sec,
            url_defined_later=url_defined_later,
        )
        self.logging.info("Initialising %s service (REST)" % self.name)
        self.last_response = None

    def http_get(self):
        # should return unicode
        raise NotImplementedError

    def http_post(self):
        raise NotImplementedError

    def http_put(self):
        raise NotImplementedError

    def http_delete(self):
        raise NotImplementedError


import requests  # replacement for urllib2 (2-3 times faster)
from requests.models import Response
import requests_cache  # use caching wihh requests

# import grequests        # use asynchronous requests with gevent
# Note that grequests should be imported after requests_cache. Otherwise,
# one should use a session instance when calling grequests.get, which we do
# here below


class REST(RESTbase):
    """

    The ideas (sync/async) and code using requests were inspired from the chembl
    python wrapper but significantly changed.

    Get one value::

        >>> from bioservices import REST
        >>> s = REST("test", "https://www.ebi.ac.uk/chemblws")
        >>> res = s.get_one("targets/CHEMBL2476.json", "json")
        >>> res['organism']
        u'Homo sapiens'

    The caching has two major interests. First one is that it speed up requests if
    you repeat requests. ::


        >>> s = REST("test", "https://www.ebi.ac.uk/chemblws")
        >>> s.CACHING = True
        >>> # requests will be stored in a local sqlite database
        >>> s.get_one("targets/CHEMBL2476")
        >>> # Disconnect your wiki and any network connections.
        >>> # Without caching you cannot fetch any requests but with
        >>> # the CACHING on, you can retrieve previous requests:
        >>> s.get_one("targets/CHEMBL2476")


    Advantages of requests over urllib

    requests length is not limited to 2000 characters
    http://www.g-loaded.eu/2008/10/24/maximum-url-length/


    There is no need for authentication if the web services available
    in bioservices except for a few exception. In such case, the username and
    password are to be provided with the method call. However,
    in the future if a services requires authentication, one can set the
    attribute :attr:`authentication` to a tuple::

        s = REST()
        s.authentication = ('user', 'pass')


    Note about headers and content type. The Accept header is
    used by HTTP clients to tell the server what content types
    they will accept. The server will then send back a
    response, which will include a Content-Type header telling
    the client what the content type of the returned content
    actually is. When using the :meth:`get__headers`, you can see
    the User-Agent, the Accept and Content-Type keys. So, here the
    HTTP requests also contain Content-Type headers. In POST or PUT requests
    the client is actually sendingdata to the server as part of the
    request, and the Content-Type header tells the server what the data actually is
    For a POST request resulting from an HTML form submission, the
    Content-Type of the request should be one of the standard form content
    types: application/x-www-form-urlencoded (default, older, simpler) or
    multipart/form-data (newer, adds support for file uploads)

    """

    content_types = {
        "bed": "text/x-bed",
        "default": "application/x-www-form-urlencoded",
        "gff3": "text/x-gff3",
        "fasta": "text/x-fasta",
        "json": "application/json",
        "jsonp": "text/javascript",
        "nh": "text/x-nh",
        "phylip": "text/x-phyloxml+xml",
        "phyloxml": "text/x-phyloxml+xml",
        "seqxml": "text/x-seqxml+xml",
        "png": "image/png",
        "jpg": "image/jpg",
        "svg": "image/svg",
        "svg+xml": "image/svg+xml",
        "gif": "image/gif",
        "jpeg": "image/jpg",
        "txt": "text/plain",
        "text": "text/plain",
        "xml": "application/xml",
        "yaml": "text/x-yaml",
    }
    # special_characters = ['/', '#', '+']

    def __init__(
        self,
        name,
        url=None,
        verbose=True,
        cache=False,
        requests_per_sec=3,
        proxies=[],
        cert=None,
        url_defined_later=False,
    ):
        super(REST, self).__init__(
            name,
            url,
            verbose=verbose,
            requests_per_sec=requests_per_sec,
            url_defined_later=url_defined_later,
        )
        self.proxies = proxies
        self.cert = cert

        bspath = self.settings.user_config_dir
        self.CACHE_NAME = bspath + os.sep + self.name + "_bioservices_db"

        self._session = None

        self.settings.params["cache.on"][0] = cache

        if self.CACHING:
            # import requests_cache
            self.logging.info("Using local cache %s" % self.CACHE_NAME)
            requests_cache.install_cache(self.CACHE_NAME)

    def delete_cache(self):
        cache_file = self.CACHE_NAME + ".sqlite"
        if os.path.exists(cache_file):
            msg = "You are about to delete this bioservices cache: %s. Proceed? (y/[n]) "
            res = input(msg % cache_file)
            if res == "y":
                os.remove(cache_file)
                self.logging.info("Removed cache")
            else:
                self.logging.info("Reply 'y' to delete the file")

    def clear_cache(self):
        from requests_cache import clear

        clear()

    def _build_url(self, query):
        url = None

        if query is None:
            url = self.url
        else:
            if query.startswith("http"):
                # assume we do want to use self.url
                url = query
            else:
                url = "%s/%s" % (self.url, query)

        return url

    def _get_session(self):
        if self._session is None:
            if self.CACHING is True:
                self._session = self._create_cache_session()
            else:
                self._session = self._create_session()
        return self._session

    session = property(_get_session)

    def _create_session(self):
        """Creates a normal session using HTTPAdapter

        max retries is defined in the :attr:`MAX_RETRIES`
        """
        self.logging.debug("Creating session (uncached version)")
        self._session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(max_retries=self.settings.MAX_RETRIES)
        # , pool_block=True does not work with asynchronous requests
        self._session.mount("http://", adapter)
        self._session.mount("https://", adapter)
        return self._session

    def _create_cache_session(self):
        """Creates a cached session using requests_cache package"""
        self.logging.debug("Creating session (cache version)")
        if not self._session:
            # import requests_cache
            self.logging.debug("No cached session created yet. Creating one")
            self._session = requests_cache.CachedSession(
                self.CACHE_NAME, backend="sqlite", fast_save=self.settings.FAST_SAVE
            )
        return self._session

    def _get_timeout(self):
        return self.settings.TIMEOUT

    def _set_timeout(self, value):
        self.settings.TIMEOUT = value

    TIMEOUT = property(_get_timeout, _set_timeout)

    def _process_get_request(self, url, session, frmt, data=None, **kwargs):
        try:
            res = session.get(url, **kwargs)
            self.last_response = res
            res = self._interpret_returned_request(res, frmt)
            return res
        except Exception:
            return None

    def _interpret_returned_request(self, res, frmt):
        # must be a Response
        if isinstance(res, Response) is False:
            return res
        # if a response, there is a status code that should be ok
        if not res.ok:
            reason = res.reason
            self.logging.warning("status is not ok with {0}".format(reason))
            return res.status_code
        if frmt == "json":
            try:
                return res.json()
            except:
                return res
        # finally
        return res.content

    def _apply(self, iterable, fn, *args, **kwargs):
        return [fn(x, *args, **kwargs) for x in iterable if x is not None]

    def _get_async(self, keys, frmt="json", params={}):
        # does not work under pyhon3 so local import
        import grequests

        session = self._get_session()
        try:
            # build the requests
            urls = self._get_all_urls(keys, frmt)
            self.logging.debug("grequests.get processing")
            rs = (grequests.get(url, session=session, params=params) for key, url in zip(keys, urls))
            # execute them
            self.logging.debug("grequests.map call")
            ret = grequests.map(rs, size=min(self.settings.CONCURRENT, len(keys)))
            self.last_response = ret
            self.logging.debug("grequests.map call done")
            return ret
        except Exception as err:
            self.logging.warning("Error caught in async. " + err.message)
            return []

    def _get_all_urls(self, keys, frmt=None):
        return ("%s/%s" % (self.url, query) for query in keys)

    def get_async(self, keys, frmt="json", params={}, **kargs):
        ret = self._get_async(keys, frmt, params=params, **kargs)
        return self._apply(ret, self._interpret_returned_request, frmt)

    def get_sync(self, keys, frmt="json", **kargs):
        return [self.get_one(key, frmt=frmt, **kargs) for key in keys]

    def http_get(self, query, frmt="json", params={}, **kargs):
        """

        * query is the suffix that will be appended to the main url attribute.
        * query is either a string or a list of strings.
        * if list is larger than ASYNC_THRESHOLD, use asynchronous call.

        """
        if isinstance(query, list) and len(query) > self.settings.ASYNC_THRESHOLD:
            self.logging.debug("Running async call for a list")
            return self.get_async(query, frmt, params=params, **kargs)

        if isinstance(query, list) and len(query) <= self.settings.ASYNC_THRESHOLD:
            self.logging.debug("Running sync call for a list")
            return [self.get_one(key, frmt, params=params, **kargs) for key in query]
            # return self.get_sync(query, frmt)

        # OTHERWISE
        self.logging.debug("Running http_get (single call mode)")
        # return self.get_one(**{'frmt': frmt, 'query': query, 'params':params})

        # if user provide a content, let us use it, otherwise, it will be the
        # same as the frmt provided
        content = kargs.get("content", self.content_types[frmt])

        # if user provide a header, we use it otherwise, we use the header from
        # bioservices and the content defined here above
        headers = kargs.get("headers")
        if headers is None:
            headers = {}
            headers["User-Agent"] = self.getUserAgent()
            if content is None:
                headers["Accept"] = self.content_types[frmt]
            else:
                headers["Accept"] = content
        kargs.update({"headers": headers})

        return self.get_one(query, frmt=frmt, params=params, **kargs)

    def get_one(self, query=None, frmt="json", params={}, **kargs):
        """

        if query starts with http:// do not use self.url
        """
        self._calls()
        url = self._build_url(query)

        if url.count("//") > 1:
            self.logging.warning("URL of the services contains a double //." + "Check your URL and remove trailing /")
        self.logging.debug(url)
        try:
            kargs["params"] = params
            kargs["timeout"] = self.TIMEOUT
            kargs["proxies"] = self.proxies
            kargs["cert"] = self.cert
            # Used only in biomart with cosmic database
            # See doc/source/biomart.rst for an example
            if hasattr(self, "authentication"):
                kargs["auth"] = self.authentication

            # res = self.session.get(url, **{'timeout':self.TIMEOUT, 'params':params})
            res = self.session.get(url, **kargs)

            self.last_response = res
            res = self._interpret_returned_request(res, frmt)
            try:
                # for python 3 compatibility
                res = res.decode()
            except:
                pass
            return res
        except Exception as err:
            self.logging.critical(err)
            self.logging.critical(
                """Query unsuccesful. Maybe too slow response.
    Consider increasing it with settings.TIMEOUT attribute {}""".format(
                    self.settings.TIMEOUT
                )
            )

    def http_post(self, query, params=None, data=None, frmt="xml", headers=None, files=None, content=None, **kargs):
        # query and frmt are bioservices parameters. Others are post parameters
        # NOTE in requests.get you can use params parameter
        # BUT in post, you use data
        # only single post implemented for now unlike get that can be asynchronous
        # or list of queries

        # if user provide a header, we use it otherwise, we use the header from
        # bioservices and the content defined here above
        if headers is None:
            headers = {}
            headers["User-Agent"] = self.getUserAgent()
            if content is None:
                headers["Accept"] = self.content_types[frmt]
            else:
                headers["Accept"] = content

        self.logging.debug("Running http_post (single call mode)")
        kargs.update({"query": query})
        kargs.update({"headers": headers})
        kargs.update({"files": files})
        kargs["proxies"] = self.proxies
        kargs["cert"] = self.cert
        kargs.update({"params": params})
        kargs.update({"data": data})
        kargs.update({"frmt": frmt})
        return self.post_one(**kargs)

    def post_one(self, query=None, frmt="json", **kargs):
        self._calls()
        self.logging.debug("BioServices:: Entering post_one function")
        if query is None:
            url = self.url
        else:
            url = "%s/%s" % (self.url, query)
        self.logging.debug(url)
        try:
            res = self.session.post(url, **kargs)
            self.last_response = res
            res = self._interpret_returned_request(res, frmt)
            try:
                return res.decode()
            except:
                self.logging.debug("BioServices:: Could not decode the response")
                return res
        except Exception as err:
            traceback.print_exc()
            return None

    def getUserAgent(self):
        # self.logging.info('getUserAgent: Begin')
        urllib_agent = "Python-requests/%s" % requests.__version__
        # clientRevision = ''
        from bioservices import version

        clientVersion = version
        user_agent = "BioServices/%s (bioservices.%s; Python %s; %s) %s" % (
            clientVersion,
            os.path.basename(__file__),
            platform.python_version(),
            platform.system(),
            urllib_agent,
        )
        # self.logging.info('getUserAgent: user_agent: ' + user_agent)
        # self.logging.info('getUserAgent: End')
        return user_agent

    def get_headers(self, content="default"):
        """
        :param str content: set to default that is application/x-www-form-urlencoded
            so that it has the same behaviour as urllib2 (Sept 2014)

        """
        headers = {}
        headers["User-Agent"] = self.getUserAgent()
        headers["Accept"] = self.content_types[content]
        headers["Content-Type"] = self.content_types[content]
        # "application/json;odata=verbose" required in reactome
        # headers['Content-Type'] = "application/json;odata=verbose" required in reactome
        return headers

    def debug_message(self):
        print(self.last_response.content)
        print(self.last_response.reason)
        print(self.last_response.status_code)

    def http_delete(self, query, params=None, frmt="xml", headers=None, **kargs):
        kargs.update({"query": query})
        kargs.update({"params": params})
        kargs.update({"frmt": frmt})

        return self.delete_one(**kargs)

    def delete_one(self, query, frmt="json", **kargs):
        self._calls()
        self.logging.debug("BioServices:: Entering delete_one function")
        if query is None:
            url = self.url
        else:
            url = "%s/%s" % (self.url, query)
        self.logging.debug(url)
        try:
            res = self.session.delete(url, **kargs)
            self.last_response = res
            res = self._interpret_returned_request(res, frmt)
            try:
                return res.decode()
            except:
                self.debug("BioServices:: Could not decode the response")
                return res
        except Exception as err:
            print(err)
            return None
        except:
            pass
